__author__ = "sibirrer"
# this file contains a class to make a gaussian

import numpy as np
from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["GaussianEllipsePotential"]


class GaussianEllipsePotential(LensProfileBase):
    """This class contains functions to evaluate a Gaussian function and calculates its
    derivative and hessian matrix with ellipticity in the convergence.

    the calculation follows Glenn van de Ven et al. 2009
    """

    param_names = ["amp", "sigma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "sigma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "sigma": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.spherical = GaussianKappa()
        self._diff = 0.000001
        super(GaussianEllipsePotential, self).__init__()

    def function(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Returns Gaussian."""

        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        x_shift = x - center_x
        y_shift = y - center_y
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)
        e = abs(1 - q)
        x_ = (cos_phi * x_shift + sin_phi * y_shift) * np.sqrt(1 - e)
        y_ = (-sin_phi * x_shift + cos_phi * y_shift) * np.sqrt(1 + e)
        f_ = self.spherical.function(x_, y_, amp=amp, sigma=sigma)
        return f_

    def derivatives(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function."""
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        x_shift = x - center_x
        y_shift = y - center_y
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)
        e = abs(1 - q)
        x_ = (cos_phi * x_shift + sin_phi * y_shift) * np.sqrt(1 - e)
        y_ = (-sin_phi * x_shift + cos_phi * y_shift) * np.sqrt(1 + e)

        f_x_prim, f_y_prim = self.spherical.derivatives(x_, y_, amp=amp, sigma=sigma)
        f_x_prim *= np.sqrt(1 - e)
        f_y_prim *= np.sqrt(1 + e)
        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        alpha_ra, alpha_dec = self.derivatives(
            x, y, amp, sigma, e1, e2, center_x, center_y
        )
        diff = self._diff
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff, y, amp, sigma, e1, e2, center_x, center_y
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff, amp, sigma, e1, e2, center_x, center_y
        )

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        f_yx = (alpha_dec_dx - alpha_dec) / diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff
        return f_xx, f_xy, f_yx, f_yy

    def density(self, r, amp, sigma, e1, e2):
        """

        :param r:
        :param amp:
        :param sigma:
        :return:
        """
        return self.spherical.density(r, amp, sigma)

    def density_2d(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param amp:
        :param sigma:
        :param e1:
        :param e2:
        :param center_x:
        :param center_y:
        :return:
        """
        return self.spherical.density_2d(x, y, amp, sigma, center_x, center_y)

    def mass_2d(self, R, amp, sigma, e1, e2):
        """

        :param R:
        :param amp:
        :param sigma:
        :param e1:
        :param e2:
        :return:
        """
        return self.spherical.mass_2d(R, amp, sigma)

    def mass_3d(self, R, amp, sigma, e1, e2):
        """

        :param R:
        :param amp:
        :param sigma:
        :param e1:
        :param e2:
        :return:
        """
        return self.spherical.mass_3d(R, amp, sigma)

    def mass_3d_lens(self, R, amp, sigma, e1, e2):
        """

        :param R:
        :param amp:
        :param sigma:
        :param e1:
        :param e2:
        :return:
        """
        return self.spherical.mass_3d_lens(R, amp, sigma)

    def mass_2d_lens(self, R, amp, sigma, e1, e2):
        """

        :param R:
        :param amp:
        :param sigma:
        :param e1:
        :param e2:
        :return:
        """
        return self.spherical.mass_2d_lens(R, amp, sigma)



from lenstronomy.GalKin.psf import PSF
import numpy as np
import numpy.testing as npt
import unittest


class TestPSF(object):
    def setup_method(self):
        pass

    def test_displace_psf(self):
        psf = PSF(psf_type="GAUSSIAN", fwhm=1)
        np.random.seed(41)
        x, y = psf.displace_psf(0, 0)
        assert x != 0
        assert y != 0

        psf = PSF(psf_type="MOFFAT", fwhm=1, moffat_beta=2.6)
        np.random.seed(41)
        x, y = psf.displace_psf(0, 0)
        assert x != 0
        assert y != 0

    def test_kernel(self):
        psf = PSF(psf_type="GAUSSIAN", fwhm=1)

        kernel = psf.convolution_kernel(delta_pix=0.3, num_pix=21)
        npt.assert_almost_equal(np.sum(kernel), 1, decimal=5)

        psf = PSF(psf_type="MOFFAT", fwhm=1, moffat_beta=2.6)
        kernel = psf.convolution_kernel(delta_pix=0.3, num_pix=21)
        npt.assert_almost_equal(np.sum(kernel), 1, decimal=5)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            psf = PSF(psf_type="BRRR", fwhm=1, moffat_beta=2.6)
            # psf.displace_psf(0, 0)



__author__ = "nataliehogg"

# man with one sampling method always knows his posterior distribution; man with two never certain.

import numpy as np


class CobayaSampler(object):
    def __init__(self, likelihood_module, mean_start, sigma_start):
        """Wrapper for pure Metropolis--Hastings MCMC sampling with Cobaya.

        If you use this sampler, you must cite the following works:

        Lewis & Bridle,
        https://arxiv.org/abs/astro-ph/0205436

        Lewis, https://arxiv.org/abs/1304.4473

        Torrado & Lewis,
        https://arxiv.org/abs/2005.05290
         and https://ascl.net/1910.019

        For more information about Cobaya, see
        https://cobaya.readthedocs.io/en/latest/index.html

        :param likelihood_module: LikelihoodModule() instance
        :param mean_start: initial point for parameters are drawn from Gaussians with
            these means
        :param sigma_start: initial point for parameters are drawn from Gaussians with
            these standard deviations
        """

        # get the logL and parameter info from LikelihoodModule
        self._check_install()
        self._likelihood_module = likelihood_module
        self._num_params, self._param_names = self._likelihood_module.param.num_param()
        (
            self._lower_limit,
            self._upper_limit,
        ) = self._likelihood_module.param.param_limits()

        self._mean_start = mean_start
        self._sigma_start = sigma_start

    def run(self, **kwargs):
        """
        :param kwargs: dictionary of keyword arguments for Cobaya. kwargs that can be passed are:
        'proposal_widths' (standard deviation of the Gaussian from which initial point is drawn, list or dict),
        'latex' (list of LaTeX lables for params),
        'path' (where products will be saved, string),
        'force_overwrite' (whether or not to overwite previous products with the same name, bool) and
        'mpi' (to run in MPI mode, bool).
        Furthermore, all the cobaya-native kwargs for the mcmc sampler listed in the docs are available: https://cobaya.readthedocs.io/en/latest/sampler_mcmc.html#options-and-defaults
        except 'drag' and 'blocking', since there is no obvious parameter speed hierarchy in a strong lensing likelihood.
        If none of these kwargs are passed, the default values/settings will be used.
        """

        sampled_params = self._param_names

        # add the priors to the sampled_params
        # currently a uniform prior is hardcoded for all params
        # cobaya allows any 1D continuous dist in scipy.stats; thinking how to implement this here
        sampled_params = {
            k: {
                "prior": {
                    "dist": "uniform",
                    "min": self._lower_limit[i],
                    "max": self._upper_limit[i],
                }
            }
            for k, i in zip(sampled_params, range(len(sampled_params)))
        }

        # add reference values to start chain close to expected best fit
        # this hardcodes a Gaussian and uses the sigma_kwargs passed by the user
        # again cobaya allows any 1D continous distribution; thinking how to implement this
        # tricky with current info internal in lenstronomy
        [
            sampled_params[k].update(
                {
                    "ref": {
                        "dist": "norm",
                        "loc": self._mean_start[i],
                        "scale": self._sigma_start[i],
                    }
                }
            )
            for k, i in zip(sampled_params.keys(), range(len(sampled_params)))
        ]

        # add proposal widths
        # first check if proposal_widths has been passed
        if "proposal_widths" not in kwargs:
            pass
        else:
            # check if what's been passed is dict
            if isinstance(kwargs["proposal_widths"], dict):
                # if yes, convert to list
                props = list(kwargs["proposal_widths"].values())
            elif isinstance(kwargs["proposal_widths"], list):
                # if no and it's a list, do nothing
                props = kwargs["proposal_widths"]
            else:
                # if no and not a list, raise TypeError
                raise TypeError(
                    "Proposal widths must be a list of floats or a dictionary of parameters and floats."
                )
            # check the right number of values are present
            if len(props) != len(sampled_params.keys()):
                # if not, raise ValueError
                raise ValueError(
                    "You must provide the same number of proposal widths as sampled parameters."
                )
            # update sampled_params dict with proposal widths
            [
                sampled_params[k].update({"proposal": props[i]})
                for k, i in zip(sampled_params.keys(), range(len(props)))
            ]

        # add LaTeX labels so lenstronomy kwarg names don't break getdist plotting
        # first check if the labels have been passed
        if "latex" not in kwargs:
            # if not, print a warning
            print(
                "No LaTeX labels provided: manually edit the updated.yaml file to avoid lenstronomy labels breaking GetDist."
            )
            pass
        else:
            latex = kwargs["latex"]
            # check the right number of labels are present
            if len(latex) != len(sampled_params.keys()):
                # if not, raise ValueError
                raise ValueError(
                    "You must provide the same number of labels as sampled parameters."
                )
            # update sampled_params dict with labels
            [
                sampled_params[k].update({"latex": latex[i]})
                for k, i in zip(sampled_params.keys(), range(len(latex)))
            ]

        def likelihood_for_cobaya(**kwargs):
            """We define a function to return the log-likelihood; this function is
            passed to Cobaya. The function must be nested within the run() function for
            it to work properly.

            :param kwargs: dictionary of keyword arguments
            """
            current_input_values = [kwargs[p] for p in sampled_params]
            logp = self._likelihood_module.likelihood(current_input_values)
            return logp

        # gather all the information to pass to cobaya, starting with the likelihood
        info = {
            "likelihood": {
                "lenstronomy_likelihood": {
                    "external": likelihood_for_cobaya,
                    "input_params": sampled_params,
                }
            }
        }

        # for the above, can we do an args2kwargs for the 'output_params' key?? might bypass plotting issue

        # parameter info
        info["params"] = sampled_params

        # get all the kwargs for the mcmc sampler in cobaya
        # if not present, passes a default value (most taken from cobaya docs)
        # note: parameter blocking and drag kwargs not provided because speed hierarchy not possible in strong lensing likelihoods

        mcmc_kwargs = {
            "burn_in": kwargs.get("burn_in", 0),
            "max_tries": kwargs.get("max_tries", 100 * self._num_params),
            "covmat": kwargs.get("covmat", None),
            "proposal_scale": kwargs.get("proposal_scale", 1),
            "output_every": kwargs.get("output_every", 500),
            "learn_every": kwargs.get("learn_every", 40 * self._num_params),
            "learn_proposal": kwargs.get("learn_proposal", True),
            "learn_proposal_Rminus1_max": kwargs.get("learn_proposal_Rminus1_max", 2),
            "learn_proposal_Rminus1_max_early": kwargs.get(
                "learn_proposal_Rminus1_max_early", 30
            ),
            "learn_proposal_Rminus1_min": kwargs.get("learn_proposal_Rminus1_min", 0),
            "max_samples": kwargs.get("max_samples", np.inf),
            "Rminus1_stop": kwargs.get("Rminus1_stop", 0.01),
            "Rminus1_cl_stop": kwargs.get("Rminus1_cl_stop", 0.2),
            "Rminus1_cl_level": kwargs.get("Rminus1_cl_level", 0.95),
            "Rminus1_single_split": kwargs.get("Rminus1_single_split", 4),
            "measure_speeds": kwargs.get("measure_speeds", True),
            "oversample_power": kwargs.get("oversample_power", 0.4),
            "oversample_thin": kwargs.get("oversample_thin", True),
        }

        if "drag" in kwargs:
            raise ValueError(
                "Parameter dragging not possible in a strong lensing likelihood."
            )

        # select mcmc as the sampler and pass the relevant kwargs
        info["sampler"] = {"mcmc": mcmc_kwargs}

        # where the chains and other files will be saved
        if "path" not in kwargs:
            info["output"] = None
        else:
            info["output"] = kwargs["path"]

        # whether or not to overwrite previous chains with the same name (bool)
        if "force_overwrite" not in kwargs:
            info["force"] = True
        else:
            info["force"] = kwargs["force_overwrite"]

        # check for mpi
        if "mpi" not in kwargs:
            kwargs["mpi"] = False

        # run the sampler
        # we wrap the call to crun to make sure any MPI exceptions are caught properly
        # this ensures the entire run will be terminated if any individual process breaks
        from cobaya.run import run as crun

        if kwargs["mpi"] == True:
            from mpi4py import MPI
            from cobaya.log import LoggedError

            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()

            success = False
            try:
                updated_info, sampler = crun(info)
                success = True
            except LoggedError as err:
                pass
            success = all(comm.allgather(success))
            if not success and rank == 0:
                print("Sampling failed!")
        else:
            comm = None  # is this necessary?
            updated_info, sampler = crun(info)

        # get the best fit (max likelihood); returns a pandas series
        # we use the native cobaya calculation instead of lenstronomy's
        # this is because crun does not directly expose the samples themselves
        best_fit_series = sampler.collection.bestfit()

        # turn that pandas series into a list (of floats)
        keys = list(sampled_params)  #  avoiding some new pandas error...
        best_fit_values = best_fit_series[keys].values.tolist()

        return updated_info, sampler, best_fit_values

    @staticmethod
    def _check_install():
        try:
            import cobaya
        except ImportError:
            print(
                "Warning : cobaya not properly installed. \
                    You can get it with $pip install cobaya."
            )
        else:
            pass



__all__ = ["AngularSensitivity"]


class AngularSensitivity(object):
    """Telescope Angular Sensitivity class. This class provides functions describing the
    EM radiation sensitivity along different directions of some specific telescopes,
    including radio antennae of an interferometric array.

    A general reference of telescope angular sensitivity can be found in Section 5.4 of
    Bradt, H. (2004). Astronomy methods: A physical approach to astronomical
    observations. Cambridge University Press.
    """

    def __init__(self, antenna_primary_beam=None):
        """

        :param antenna_primary_beam: 2d numpy array.
         Primary beam is the angular power sensitivity of EM radiation antennae of some specific telescopes.
         Usually the radiation sensitivity is largest at the center of the Field of View (FOV), and decays as the distance increases from the center.
         If the primary beam applies, it should be multiplied on the unconvolved model images, as regions with less EM sensitivity
         get less flux from the model image after the multiplication with the corresponding primary beam.
         For interferometric users, the primary beam should be provided by data reduction softwares like CASA.
        """
        self._pb = antenna_primary_beam

    @property
    def primary_beam(self):
        """
        :return: 2d numpy array of primary beam
        """
        return self._pb



from lenstronomy.SimulationAPI.observation_api import (
    Instrument,
    Observation,
    SingleBand,
)
import lenstronomy.Util.util as util
import numpy.testing as npt
import numpy as np
import unittest


class TestInstrumentObservation(object):
    def setup_method(self):
        pass

    def test_instrument(self):
        ccd_gain = 4
        pixel_scale = 0.13
        read_noise = 10
        kwargs_instrument = {
            "read_noise": read_noise,
            "pixel_scale": pixel_scale,
            "ccd_gain": ccd_gain,
        }
        instrument = Instrument(**kwargs_instrument)
        assert instrument.ccd_gain == ccd_gain
        assert instrument.pixel_scale == pixel_scale

    def test_observations(self):
        exposure_time = 90
        sky_brightness = 20.0
        num_exposures = (2,)
        seeing = 0.9
        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "GAUSSIAN",
        }
        observation = Observation(**kwargs_observations)
        assert observation.exposure_time == exposure_time * num_exposures

        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "NONE",
        }
        observation = Observation(**kwargs_observations)
        assert observation.exposure_time == exposure_time * num_exposures

    def test_update_observation(self):
        exposure_time = 90
        sky_brightness = 20.0
        num_exposures = 2
        seeing = 0.9
        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "GAUSSIAN",
        }
        observation = Observation(**kwargs_observations)

        exposure_time = 1
        sky_brightness = 1.0
        num_exposures = 1
        seeing = 1
        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "GAUSSIAN",
            "kernel_point_source": 1,
        }
        observation.update_observation(**kwargs_observations)
        assert observation.exposure_time == 1
        psf = observation.psf_class
        assert psf.fwhm == 1

    def test_psf_class(self):
        kwargs_observations = {
            "exposure_time": 1,
            "sky_brightness": 1,
            "num_exposures": 1,
            "psf_type": "NONE",
        }
        observation = Observation(**kwargs_observations)
        psf_class = observation.psf_class
        assert psf_class.psf_type == "NONE"


class TestRaise(unittest.TestCase):
    def test_raise(self):
        self.ccd_gain = 4.0
        pixel_scale = 0.13
        self.read_noise = 10.0
        kwargs_instrument = {
            "read_noise": self.read_noise,
            "pixel_scale": pixel_scale,
            "ccd_gain": self.ccd_gain,
        }

        exposure_time = 100
        sky_brightness = 20.0
        self.magnitude_zero_point = 21.0
        num_exposures = 2
        seeing = 0.9
        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "magnitude_zero_point": self.magnitude_zero_point,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "GAUSSIAN",
        }
        self.kwargs_data = util.merge_dicts(kwargs_instrument, kwargs_observations)
        with self.assertRaises(ValueError):
            SingleBand(data_count_unit="wrong", **self.kwargs_data)

        with self.assertRaises(ValueError):
            band = SingleBand(
                pixel_scale=1,
                exposure_time=1,
                magnitude_zero_point=1,
                read_noise=None,
                ccd_gain=None,
                sky_brightness=None,
                seeing=None,
                num_exposures=1,
                psf_type="GAUSSIAN",
                kernel_point_source=None,
                data_count_unit="ADU",
                background_noise=None,
            )
            out = band.sky_brightness

        with self.assertRaises(ValueError):
            band = SingleBand(
                pixel_scale=1,
                exposure_time=1,
                magnitude_zero_point=1,
                read_noise=None,
                ccd_gain=None,
                sky_brightness=None,
                seeing=None,
                num_exposures=1,
                psf_type="GAUSSIAN",
                kernel_point_source=None,
                data_count_unit="ADU",
                background_noise=None,
            )
            out = band.background_noise


class TestData(object):
    def setup_method(self):
        self.ccd_gain = 4.0
        pixel_scale = 0.13
        self.read_noise = 10.0
        self.kwargs_instrument = {
            "read_noise": self.read_noise,
            "pixel_scale": pixel_scale,
            "ccd_gain": self.ccd_gain,
        }

        exposure_time = 100
        sky_brightness = 20.0
        self.magnitude_zero_point = 21.0
        num_exposures = 2
        seeing = 0.9
        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "magnitude_zero_point": self.magnitude_zero_point,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "GAUSSIAN",
        }
        self.kwargs_data = util.merge_dicts(self.kwargs_instrument, kwargs_observations)
        self.data_adu = SingleBand(data_count_unit="ADU", **self.kwargs_data)
        self.data_e_ = SingleBand(data_count_unit="e-", **self.kwargs_data)

    def test_sky_brightness(self):
        sky_adu = self.data_adu.sky_brightness
        sky_e_ = self.data_e_.sky_brightness
        assert sky_e_ == sky_adu * self.ccd_gain
        npt.assert_almost_equal(sky_adu, 0.627971607877395, decimal=6)

    def test_background_noise(self):
        bkg_adu = self.data_adu.background_noise
        bkg_e_ = self.data_e_.background_noise
        assert bkg_adu == bkg_e_ / self.ccd_gain

        self.data_adu._background_noise = 1
        bkg = self.data_adu.background_noise
        assert bkg == 1

    def test_flux_noise(self):
        flux_iid = 50.0
        flux_adu = flux_iid / self.ccd_gain
        noise_adu = self.data_adu.flux_noise(flux_adu)
        noise_e_ = self.data_e_.flux_noise(flux_iid)
        assert noise_e_ == 100.0 / 200.0
        assert noise_e_ == noise_adu * self.ccd_gain

    def test_noise_for_model(self):
        model_adu = np.ones((10, 10))
        model_e_ = model_adu * self.ccd_gain
        noise_adu = self.data_adu.noise_for_model(
            model_adu, background_noise=True, poisson_noise=True, seed=42
        )
        noise_adu_2 = self.data_adu.noise_for_model(
            model_adu, background_noise=True, poisson_noise=True, seed=42
        )
        npt.assert_almost_equal(noise_adu, noise_adu_2, decimal=10)
        noise_e_ = self.data_e_.noise_for_model(
            model_e_, background_noise=True, poisson_noise=True, seed=42
        )
        npt.assert_almost_equal(noise_adu, noise_e_ / self.ccd_gain, decimal=10)
        noise_e_ = self.data_e_.noise_for_model(
            model_e_, background_noise=True, poisson_noise=True, seed=None
        )

    def test_estimate_noise(self):
        image_adu = np.ones((10, 10))
        image_e_ = image_adu * self.ccd_gain
        noise_adu = self.data_adu.estimate_noise(image_adu)
        noise_e_ = self.data_e_.estimate_noise(image_e_)
        npt.assert_almost_equal(noise_e_, noise_adu * self.ccd_gain)

    def test_magnitude2cps(self):
        mag_0 = self.data_adu.magnitude2cps(magnitude=self.magnitude_zero_point)
        npt.assert_almost_equal(mag_0, 1.0 / self.ccd_gain, decimal=10)
        mag_0_e_ = self.data_e_.magnitude2cps(magnitude=self.magnitude_zero_point)
        npt.assert_almost_equal(mag_0_e_, 1, decimal=10)

        mag_0 = self.data_adu.magnitude2cps(magnitude=self.magnitude_zero_point + 1)
        npt.assert_almost_equal(mag_0, 0.0995267926383743, decimal=10)

        mag_0 = self.data_adu.magnitude2cps(magnitude=self.magnitude_zero_point - 1)
        npt.assert_almost_equal(mag_0, 0.627971607877395, decimal=10)

    def test_flux_iid(self):
        flux_iid_adu = self.data_adu.flux_iid(flux_per_second=1)
        flux_iid_e = self.data_e_.flux_iid(flux_per_second=1)
        npt.assert_almost_equal(flux_iid_e, flux_iid_adu / self.ccd_gain, decimal=6)

        flux_adu = 10
        flux_e_ = flux_adu * self.ccd_gain
        noise_e_ = self.data_e_.flux_noise(flux_e_)
        noise_adu = self.data_adu.flux_noise(flux_adu)
        npt.assert_almost_equal(noise_e_ / self.ccd_gain, noise_adu, decimal=8)

    def test_psf_type(self):
        assert self.data_adu._psf_type == "GAUSSIAN"
        kwargs_observations = {
            "exposure_time": 1,
            "sky_brightness": 1,
            "magnitude_zero_point": self.magnitude_zero_point,
            "num_exposures": 1,
            "seeing": 1,
            "psf_type": "PIXEL",
        }
        kwargs_data = util.merge_dicts(self.kwargs_instrument, kwargs_observations)
        data_pixel = SingleBand(data_count_unit="ADU", **kwargs_data)
        assert data_pixel._psf_type == "PIXEL"

        kwargs_observations = {
            "exposure_time": 1,
            "sky_brightness": 1,
            "magnitude_zero_point": self.magnitude_zero_point,
            "num_exposures": 1,
            "seeing": 1,
            "psf_type": "NONE",
        }
        kwargs_data = util.merge_dicts(self.kwargs_instrument, kwargs_observations)
        data_pixel = SingleBand(data_count_unit="ADU", **kwargs_data)
        assert data_pixel._psf_type == "NONE"



__author__ = "nataliehogg"

import pytest
import numpy as np
from unittest import TestCase

from lenstronomy.Sampling.Samplers.cobaya_sampler import CobayaSampler


@pytest.fixture
def import_fixture(simple_einstein_ring_likelihood):
    """
    :param simple_einstein_ring_likelihood_2d: fixture
    :return:
    """
    likelihood, kwargs_truths = simple_einstein_ring_likelihood
    means = likelihood.param.kwargs2args(**kwargs_truths)
    sigmas = np.ones_like(means) * 0.1
    sampler = CobayaSampler(
        likelihood_module=likelihood, mean_start=means, sigma_start=sigmas
    )
    return sampler, likelihood, means, sigmas


class TestCobayaSampler(object):
    """Test cobaya."""

    def setup_method(self):
        pass

    def test_sampler(self, import_fixture):
        """Function to test the sampler."""
        # test the sampler
        sampler, likelihood, means, sigmas = import_fixture

        test_cobaya = {"Rminus1_stop": 100}

        updated_info, sampler_name, best_fit_values = sampler.run(**test_cobaya)

        assert str(sampler_name) == "mcmc"

        # test labels
        sampler, likelihood, means, sigmas = import_fixture

        test_labels_kwargs = {"Rminus1_stop": 100, "latex": ["theta_{\rm E}"]}

        updated_info_l, sampler_name_l, best_fit_values_l = sampler.run(
            **test_labels_kwargs
        )

        assert str(sampler_name_l) == "mcmc"

        # test passing dict for proposals
        sampler, likelihood, means, sigmas = import_fixture

        props = {"theta_E": 0.001}

        test_prop_kwargs = {"Rminus1_stop": 100, "proposal_widths": props}

        updated_info_d, sampler_name_d, best_fit_values_d = sampler.run(
            **test_prop_kwargs
        )

        assert str(sampler_name_d) == "mcmc"

        # test passing path
        sampler, likelihood, means, sigmas = import_fixture

        test_path_kwargs = {"Rminus1_stop": 100, "path": "test_chain"}

        updated_info_p, sampler_name_p, best_fit_values_p = sampler.run(
            **test_path_kwargs
        )

        assert str(sampler_name_p) == "mcmc"

        # use unittest to test raised exceptions
        t = TestCase()

        with t.assertRaises(TypeError):
            # checks that TypeError is raised if prop widths not list or dict
            test_prop_type = {"proposal_widths": 0.1}
            sampler.run(**test_prop_type)

        with t.assertRaises(ValueError):
            # checks that ValueError is raised if wrong number of prop width
            test_prop_num = {"proposal_widths": [0.1, 0.1, 0.1]}
            sampler.run(**test_prop_num)

        with t.assertRaises(ValueError):
            # checks that ValueError is raised if wrong number of labels
            test_latex_num = {"latex": ["theta_{\rm E}", "gamma"]}
            sampler.run(**test_latex_num)

        with t.assertRaises(ValueError):
            # checks that ValueError is raised if drag is passed
            test_drag = {"drag": True}
            sampler.run(**test_drag)


if __name__ == "__main__":
    pytest.main()



__author__ = "dgilman"

import unittest
from lenstronomy.LensModel.Profiles.general_nfw import GNFW
from lenstronomy.LensModel.lens_model import LensModel
from scipy.integrate import quad
from lenstronomy.LensModel.Profiles.splcore import SPLCORE

import numpy.testing as npt
import pytest


class TestGNFW(object):
    def setup_method(self):
        self.gnfw = GNFW()
        self.splcore = SPLCORE()
        self.kwargs_lens = {
            "alpha_Rs": 2.1,
            "Rs": 1.5,
            "gamma_inner": 1.0,
            "gamma_outer": 3.0,
            "center_x": 0.04,
            "center_y": -1.0,
        }

    def test_alphaRs(self):
        alpha_rs = self.gnfw.derivatives(
            self.kwargs_lens["Rs"],
            0.0,
            self.kwargs_lens["Rs"],
            self.kwargs_lens["alpha_Rs"],
            self.kwargs_lens["gamma_inner"],
            self.kwargs_lens["gamma_outer"],
        )[0]
        npt.assert_almost_equal(alpha_rs, self.kwargs_lens["alpha_Rs"], 8)

    def test_alphaRs_rho0_conversion(self):
        rho0 = self.gnfw.alpha2rho0(
            self.kwargs_lens["alpha_Rs"],
            self.kwargs_lens["Rs"],
            self.kwargs_lens["gamma_inner"],
            self.kwargs_lens["gamma_outer"],
        )
        alpha_Rs = self.gnfw.rho02alpha(
            rho0,
            self.kwargs_lens["Rs"],
            self.kwargs_lens["gamma_inner"],
            self.kwargs_lens["gamma_outer"],
        )
        npt.assert_almost_equal(alpha_Rs, self.kwargs_lens["alpha_Rs"], 5)

    def test_lensing_quantities(self):
        lensmodel = LensModel(["GNFW"])
        f_x, f_y = self.gnfw.derivatives(1.0, 1.5, **self.kwargs_lens)
        f_x_, f_y_ = lensmodel.alpha(1.0, 1.5, [self.kwargs_lens])
        npt.assert_almost_equal(f_x, f_x_, 5)
        npt.assert_almost_equal(f_y, f_y_, 5)

        f_xx, f_xy, f_yx, f_yy = self.gnfw.hessian(1.0, 1.5, **self.kwargs_lens)
        f_xx_, f_xy_, f_yx_, f_yy_ = lensmodel.hessian(1.0, 1.5, [self.kwargs_lens])
        npt.assert_almost_equal(f_xx, f_xx_, 5)
        npt.assert_almost_equal(f_yy, f_yy_, 5)
        npt.assert_almost_equal(f_xy, f_xy_, 5)

    def test_mass2d(self):
        rho0 = self.gnfw.alpha2rho0(
            self.kwargs_lens["alpha_Rs"],
            self.kwargs_lens["Rs"],
            self.kwargs_lens["gamma_inner"],
            self.kwargs_lens["gamma_outer"],
        )
        m2d = self.gnfw.mass_2d(
            10.0,
            self.kwargs_lens["Rs"],
            rho0,
            self.kwargs_lens["gamma_inner"],
            self.kwargs_lens["gamma_outer"],
        )
        integrand = (
            lambda x: 2
            * 3.14159265
            * x
            * self.gnfw.density_2d(
                x,
                0.0,
                self.kwargs_lens["Rs"],
                rho0,
                self.kwargs_lens["gamma_inner"],
                self.kwargs_lens["gamma_outer"],
            )
        )
        m2d_num = quad(integrand, 0, 10.0)[0]
        npt.assert_almost_equal(m2d_num / m2d, 1.0, 5)

    def test_spl_core_match(self):
        rs = 1.5
        kwargs_spl = {"sigma0": 1e13, "gamma": 3.0, "r_core": 0.00000001}
        alpha_rs = self.splcore.derivatives(rs, 0.0, **kwargs_spl)[0]
        kwargs_gnfw = {
            "alpha_Rs": alpha_rs,
            "Rs": rs,
            "gamma_inner": 2.99999,
            "gamma_outer": 3.00001,
        }
        m3d_gnfw = self.gnfw.mass_3d_lens(5 * rs, **kwargs_gnfw)
        m3d_splcore = self.splcore.mass_3d_lens(5 * rs, **kwargs_spl)
        npt.assert_almost_equal(m3d_gnfw / m3d_splcore, 0.935, 3)
        # approximate match to splcore with similar properties


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.LensModel.Solver.solver4point import Solver4Point
from lenstronomy.LensModel.lens_model import LensModel
import lenstronomy.Util.param_util as param_util


class TestSolver4Point(object):
    def setup_method(self):
        """

        :return:
        """
        pass

    def test_decoupling(self):
        lens_model_list = ["SPEP", "SIS"]
        lensModel = LensModel(lens_model_list)
        solver = Solver4Point(lensModel)
        solver_decoupled = Solver4Point(lensModel)
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        deltapix = 0.05
        numPix = 150
        gamma = 1.9
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"theta_E": 0.1, "center_x": 0.5, "center_y": 0},
        ]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        phi_G, q = 1.5, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens_init = [
            {
                "theta_E": 1.3,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"theta_E": 0.1, "center_x": 0.5, "center_y": 0},
        ]
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        kwargs_lens_new_2, accuracy = solver_decoupled.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        print(kwargs_lens_new_2)
        print(kwargs_lens_new)
        npt.assert_almost_equal(
            kwargs_lens_new[0]["theta_E"], kwargs_lens[0]["theta_E"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e1"], kwargs_lens[0]["e1"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e2"], kwargs_lens[0]["e2"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_x"], kwargs_lens[0]["center_x"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_y"], kwargs_lens[0]["center_y"], decimal=3
        )

        npt.assert_almost_equal(
            kwargs_lens_new[0]["theta_E"], kwargs_lens_new_2[0]["theta_E"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e1"], kwargs_lens_new_2[0]["e1"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e2"], kwargs_lens_new_2[0]["e2"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_x"], kwargs_lens_new_2[0]["center_x"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_y"], kwargs_lens_new_2[0]["center_y"], decimal=3
        )

        npt.assert_almost_equal(kwargs_lens_new[0]["theta_E"], 1.0, decimal=3)
        lensModel = LensModel(lens_model_list=lens_model_list)
        x_source_new, y_source_new = lensModel.ray_shooting(
            x_pos, y_pos, kwargs_lens_new
        )
        dist = np.sqrt(
            (x_source_new - x_source_new[0]) ** 2
            + (y_source_new - y_source_new[0]) ** 2
        )
        print(dist)
        assert np.max(dist) < 0.000001

    def test_solver_spep(self):
        lens_model_list = ["SPEP"]
        lensModel = LensModel(lens_model_list)
        solver = Solver4Point(lensModel)
        lensEquationSolver = LensEquationSolver(lensModel)

        sourcePos_x = 0.1
        sourcePos_y = -0.1
        deltapix = 0.05
        numPix = 150
        gamma = 1.9
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0.1,
                "center_y": -0.1,
            }
        ]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        phi_G, q = 1.5, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens_init = [
            {
                "theta_E": 1.3,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0.0,
                "center_y": 0,
            }
        ]
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["theta_E"], kwargs_lens[0]["theta_E"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e1"], kwargs_lens[0]["e1"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e2"], kwargs_lens[0]["e2"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_x"], kwargs_lens[0]["center_x"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_y"], kwargs_lens[0]["center_y"], decimal=3
        )
        npt.assert_almost_equal(kwargs_lens_new[0]["theta_E"], 1.0, decimal=3)
        lensModel = LensModel(lens_model_list=lens_model_list)
        x_source_new, y_source_new = lensModel.ray_shooting(
            x_pos, y_pos, kwargs_lens_new
        )
        dist = np.sqrt(
            (x_source_new - x_source_new[0]) ** 2
            + (y_source_new - y_source_new[0]) ** 2
        )
        print(dist)
        assert np.max(dist) < 0.000001

    def test_solver_nfw(self):
        lens_model_list = ["NFW_ELLIPSE", "SIS"]
        lensModel = LensModel(lens_model_list)
        solver = Solver4Point(lensModel)
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        deltapix = 0.05
        numPix = 150
        Rs = 4.0
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "alpha_Rs": 1.0,
                "Rs": Rs,
                "e1": e1,
                "e2": e2,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"theta_E": 1, "center_x": 0, "center_y": 0},
        ]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        phi_G, q = 1.5, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens_init = [
            {
                "alpha_Rs": 0.5,
                "Rs": Rs,
                "e1": e1,
                "e2": e2,
                "center_x": 0.0,
                "center_y": 0,
            },
            kwargs_lens[1],
        ]
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["alpha_Rs"], kwargs_lens[0]["alpha_Rs"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e1"], kwargs_lens[0]["e1"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e2"], kwargs_lens[0]["e2"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_x"], kwargs_lens[0]["center_x"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_y"], kwargs_lens[0]["center_y"], decimal=3
        )

    def test_solver_shapelets(self):
        lens_model_list = ["SHAPELETS_CART", "SPEP"]
        lensModel = LensModel(lens_model_list)
        solver = Solver4Point(lensModel)
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.0
        deltapix = 0.05
        numPix = 150
        coeffs = np.array([0, 0.1, 0.1, 0, 0, -0.1])
        kwargs_lens = [
            {"beta": 1.0, "coeffs": coeffs, "center_x": 0.0, "center_y": 0.0},
            {
                "theta_E": 1.0,
                "gamma": 2,
                "e1": 0.1,
                "e2": 0,
                "center_x": 0,
                "center_y": 0,
            },
        ]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, y_pos)
        kwargs_lens_init = [
            {
                "beta": 1,
                "coeffs": np.zeros_like(coeffs),
                "center_x": 0.0,
                "center_y": 0,
            },
            kwargs_lens[1],
        ]
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["beta"], kwargs_lens[0]["beta"], decimal=3
        )
        coeffs_new = kwargs_lens_new[0]["coeffs"]
        for i in range(len(coeffs)):
            npt.assert_almost_equal(coeffs_new[i], coeffs[i], decimal=3)

    def test_solver_simplified(self):
        lens_model_list = ["SPEP", "SHEAR_GAMMA_PSI"]
        lensModel = LensModel(lens_model_list)

        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        deltapix = 0.05
        numPix = 150
        gamma = 1.9
        gamma_ext = 0.05
        psi_ext = 0.4
        # e1, e2 = param_util.phi_gamma_ellipticity(phi=psi_ext, gamma=gamma_ext)
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": 0.1,
                "e2": -0.1,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"gamma_ext": gamma_ext, "psi_ext": psi_ext},
        ]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        e1_new, e2_new = param_util.shear_polar2cartesian(
            phi=0.0, gamma=gamma_ext + 0.1
        )
        kwargs_lens_init = [
            {
                "theta_E": 1.3,
                "gamma": gamma,
                "e1": 0.0,
                "e2": 0.0,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"gamma_ext": gamma_ext + 0.1, "psi_ext": 0},
        ]
        solver = Solver4Point(lensModel, solver_type="PROFILE_SHEAR")
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        assert accuracy < 10 ** (-10)
        x_source, y_source = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens_new)
        x_source, y_source = np.mean(x_source), np.mean(y_source)
        x_pos_new, y_pos_new = lensEquationSolver.findBrightImage(
            x_source,
            y_source,
            kwargs_lens_new,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, x_pos_new)
        x_pos = np.sort(x_pos)
        x_pos_new = np.sort(x_pos_new)
        y_pos = np.sort(y_pos)
        y_pos_new = np.sort(y_pos_new)
        for i in range(len(x_pos)):
            npt.assert_almost_equal(x_pos[i], x_pos_new[i], decimal=6)
            npt.assert_almost_equal(y_pos[i], y_pos_new[i], decimal=6)

    def test_solver_simplified_2(self):
        lens_model_list = ["SPEP", "SHEAR_GAMMA_PSI"]
        lensModel = LensModel(lens_model_list)

        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        deltapix = 0.05
        numPix = 150
        gamma = 1.96
        e1, e2 = -0.01, -0.01
        psi_ext, gamma_ext = param_util.shear_cartesian2polar(e1, e2)
        kwargs_shear = {
            "gamma_ext": gamma_ext,
            "psi_ext": psi_ext,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": gamma,
            "center_x": 0,
            "center_y": 0,
            "e1": -0.2,
            "e2": -0.03,
        }
        kwargs_lens = [kwargs_spemd, kwargs_shear]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        kwargs_lens_init = [
            {
                "theta_E": 1.3,
                "gamma": gamma,
                "e1": 0,
                "e2": 0,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"gamma_ext": gamma_ext, "psi_ext": psi_ext},
        ]
        solver = Solver4Point(lensModel, solver_type="PROFILE_SHEAR")
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        assert accuracy < 10 ** (-10)
        x_source, y_source = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens_new)
        x_source, y_source = np.mean(x_source), np.mean(y_source)
        x_pos_new, y_pos_new = lensEquationSolver.findBrightImage(
            x_source,
            y_source,
            kwargs_lens_new,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, x_pos_new)
        x_pos = np.sort(x_pos)
        x_pos_new = np.sort(x_pos_new)
        y_pos = np.sort(y_pos)
        y_pos_new = np.sort(y_pos_new)
        for i in range(len(x_pos)):
            npt.assert_almost_equal(x_pos[i], x_pos_new[i], decimal=6)
            npt.assert_almost_equal(y_pos[i], y_pos_new[i], decimal=6)
        npt.assert_almost_equal(
            kwargs_lens_new[1]["psi_ext"], kwargs_lens[1]["psi_ext"], decimal=8
        )
        npt.assert_almost_equal(
            kwargs_lens_new[1]["gamma_ext"], kwargs_lens[1]["gamma_ext"], decimal=8
        )

    def test_solver_profile_shear(self):
        lens_model_list = ["SPEP", "SHEAR_GAMMA_PSI"]
        lensModel = LensModel(lens_model_list)

        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.0
        sourcePos_y = 0.1
        deltapix = 0.05
        numPix = 150
        gamma = 1.98
        e1, e2 = -0.04, -0.01
        gamma_ext = np.sqrt(e1**2 + e2**2)
        kwargs_shear = {
            "gamma_ext": gamma_ext,
            "psi_ext": 0.3,
        }  # shear values to the source plane
        kwargs_spemd = {
            "theta_E": 1.66,
            "gamma": gamma,
            "center_x": 0.0,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.05,
        }  # parameters of the deflector lens model

        kwargs_lens = [kwargs_spemd, kwargs_shear]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, y_pos, "test positions")
        kwargs_lens_init = [
            {
                "theta_E": 1.3,
                "gamma": gamma,
                "e1": 0,
                "e2": 0,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"gamma_ext": gamma_ext, "psi_ext": -0.3},
        ]
        solver = Solver4Point(lensModel, solver_type="PROFILE_SHEAR")
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        assert accuracy < 10 ** (-10)
        x_source, y_source = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens_new)
        x_source, y_source = np.mean(x_source), np.mean(y_source)
        x_pos_new, y_pos_new = lensEquationSolver.findBrightImage(
            x_source,
            y_source,
            kwargs_lens_new,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, x_pos_new)
        x_pos = np.sort(x_pos)
        x_pos_new = np.sort(x_pos_new)
        y_pos = np.sort(y_pos)
        y_pos_new = np.sort(y_pos_new)
        for i in range(len(x_pos)):
            npt.assert_almost_equal(x_pos[i], x_pos_new[i], decimal=6)
            npt.assert_almost_equal(y_pos[i], y_pos_new[i], decimal=6)
        npt.assert_almost_equal(
            kwargs_lens_new[1]["gamma_ext"], kwargs_lens[1]["gamma_ext"], decimal=8
        )
        npt.assert_almost_equal(
            kwargs_lens_new[1]["psi_ext"], kwargs_lens[1]["psi_ext"], decimal=8
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e1"], kwargs_lens[0]["e1"], decimal=8
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e2"], kwargs_lens[0]["e2"], decimal=8
        )

    def test_solver_profile_shear_2(self):
        lens_model_list = ["SPEP", "SHEAR"]
        lensModel = LensModel(lens_model_list)

        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.0
        sourcePos_y = 0.1
        deltapix = 0.05
        numPix = 150
        gamma = 1.98
        gamma1, gamma2 = -0.04, -0.01

        kwargs_shear = {
            "gamma1": gamma1,
            "gamma2": gamma2,
        }  # shear values to the source plane
        kwargs_spemd = {
            "theta_E": 1.66,
            "gamma": gamma,
            "center_x": 0.0,
            "center_y": 0.0,
            "e1": 0.1,
            "e2": 0.05,
        }  # parameters of the deflector lens model

        kwargs_lens = [kwargs_spemd, kwargs_shear]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, y_pos, "test positions")

        gamma_ext = np.sqrt(gamma1**2 + gamma2**2)
        gamma1_init, gamma2_init = param_util.shear_polar2cartesian(
            gamma=gamma_ext, phi=-1.3
        )

        kwargs_lens_init = [
            {
                "theta_E": 1.3,
                "gamma": gamma,
                "e1": 0,
                "e2": 0,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"gamma1": gamma1_init, "gamma2": gamma2_init},
        ]
        solver = Solver4Point(lensModel, solver_type="PROFILE_SHEAR")
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        assert accuracy < 10 ** (-10)
        x_source, y_source = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens_new)
        x_source, y_source = np.mean(x_source), np.mean(y_source)
        x_pos_new, y_pos_new = lensEquationSolver.findBrightImage(
            x_source,
            y_source,
            kwargs_lens_new,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, x_pos_new)
        x_pos = np.sort(x_pos)
        x_pos_new = np.sort(x_pos_new)
        y_pos = np.sort(y_pos)
        y_pos_new = np.sort(y_pos_new)
        for i in range(len(x_pos)):
            npt.assert_almost_equal(x_pos[i], x_pos_new[i], decimal=6)
            npt.assert_almost_equal(y_pos[i], y_pos_new[i], decimal=6)
        npt.assert_almost_equal(
            kwargs_lens_new[1]["gamma1"], kwargs_lens[1]["gamma1"], decimal=8
        )
        npt.assert_almost_equal(
            kwargs_lens_new[1]["gamma2"], kwargs_lens[1]["gamma2"], decimal=8
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e1"], kwargs_lens[0]["e1"], decimal=8
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e2"], kwargs_lens[0]["e2"], decimal=8
        )

    def test_solver_multiplane(self):
        lens_model_list = ["SPEP", "SHEAR", "SIS"]
        lensModel = LensModel(
            lens_model_list,
            z_source=1,
            lens_redshift_list=[0.5, 0.5, 0.3],
            multi_plane=True,
        )

        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.01
        sourcePos_y = -0.01
        deltapix = 0.05
        numPix = 150
        gamma = 1.96
        gamma1, gamma2 = 0.01, 0.01
        kwargs_shear = {
            "gamma1": gamma1,
            "gamma2": gamma2,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": gamma,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.2,
            "e2": 0.03,
        }
        kwargs_sis = {"theta_E": 0.1, "center_x": 1, "center_y": 0}
        kwargs_lens = [kwargs_spemd, kwargs_shear, kwargs_sis]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, y_pos, "test positions")
        kwargs_lens_init = [
            {
                "theta_E": 1.3,
                "gamma": gamma,
                "e1": 0.1,
                "e2": 0,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"gamma1": gamma1, "gamma2": gamma2},
            {"theta_E": 0.1, "center_x": 1, "center_y": 0},
        ]
        solver = Solver4Point(lensModel, solver_type="PROFILE")
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )
        print(kwargs_lens_new, "kwargs_lens_new")
        assert accuracy < 10 ** (-10)
        x_source, y_source = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens_new)
        x_source, y_source = np.mean(x_source), np.mean(y_source)
        x_pos_new, y_pos_new = lensEquationSolver.findBrightImage(
            x_source,
            y_source,
            kwargs_lens_new,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(x_pos, x_pos_new)
        x_pos = np.sort(x_pos)
        x_pos_new = np.sort(x_pos_new)
        y_pos = np.sort(y_pos)
        y_pos_new = np.sort(y_pos_new)
        for i in range(len(x_pos)):
            npt.assert_almost_equal(x_pos[i], x_pos_new[i], decimal=6)
            npt.assert_almost_equal(y_pos[i], y_pos_new[i], decimal=6)
        npt.assert_almost_equal(
            kwargs_lens_new[1]["gamma1"], kwargs_lens[1]["gamma1"], decimal=8
        )


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import lenstronomy.Util.util as util

import numpy as np
import pytest
import numpy.testing as npt
import unittest


def test_estimate_theta_E():
    x = np.array([-0.45328229, 0.57461556, 0.53757501, -0.42312438])
    y = np.array([0.69582971, -0.51226356, 0.37577509, -0.40245467])

    approx = util.approx_theta_E(x, y)
    npt.assert_array_less(approx - 1, 0.2)


def test_sort_img_index():
    ximg, yimg = np.array([1, 2, 3, 4]), np.array([0, 0, 1, 2])

    xref, yref = np.array([2, 3, 1, 4]), np.array([0, 1, 0, 2])

    indexes = util.sort_image_index(ximg, yimg, xref, yref)

    npt.assert_allclose(ximg[indexes], xref)
    npt.assert_allclose(yimg[indexes], yref)

    xref, yref = np.array([2, 3, 1, 4]), np.array([0, 1, 0, 2])

    indexes = util.sort_image_index(xref, yref, xref, yref)

    npt.assert_allclose(np.array(indexes), [0, 1, 2, 3])


def test_map_coord2pix():
    ra = 0
    dec = 0
    x_0 = 1
    y_0 = -1
    M = np.array([[1, 0], [0, 1]])
    x, y = util.map_coord2pix(ra, dec, x_0, y_0, M)
    assert x == 1
    assert y == -1

    ra = [0, 1, 2]
    dec = [0, 2, 1]
    x, y = util.map_coord2pix(ra, dec, x_0, y_0, M)
    assert x[0] == 1
    assert y[0] == -1
    assert x[1] == 2

    M = np.array([[0, 1], [1, 0]])
    x, y = util.map_coord2pix(ra, dec, x_0, y_0, M)
    assert x[1] == 3
    assert y[1] == 0


def test_make_grid():
    numPix = 11
    deltapix = 1.0

    grid = util.make_grid(numPix, deltapix)
    assert grid[0][0] == -5
    assert np.sum(grid[0]) == 0.0

    x_grid, y_grid = util.make_grid(numPix, deltapix, subgrid_res=2.0)
    assert np.sum(x_grid) == 0.0
    assert x_grid[0] == -5.25

    x_grid, y_grid = util.make_grid(numPix, deltapix, subgrid_res=1, left_lower=True)
    assert x_grid[0] == 0.0
    assert y_grid[0] == 0.0

    # Similar tests for a non-rectangular grid

    x_grid, y_grid = util.make_grid((numPix, numPix - 1), deltapix)
    assert x_grid[0] == -5.0
    assert y_grid[0] == -4.5
    assert np.sum(x_grid) == np.sum(y_grid) == 0

    x_grid, y_grid = util.make_grid(numPix, deltapix, subgrid_res=2.0)
    assert np.sum(x_grid) == np.sum(y_grid) == 0
    assert x_grid[0] == -5.25

    x_grid, y_grid = util.make_grid(numPix, deltapix, left_lower=True)
    assert x_grid[0] == 0
    assert y_grid[0] == 0


def test_make_grid_transform():
    numPix = 11
    theta = np.pi / 2
    deltaPix = 0.05
    Mpix2coord = (
        np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
        * deltaPix
    )
    ra_coord, dec_coord = util.make_grid_transformed(numPix, Mpix2coord)
    ra2d = util.array2image(ra_coord)
    assert ra2d[5, 5] == 0
    assert ra2d[4, 5] == deltaPix
    npt.assert_almost_equal(ra2d[5, 4], 0, decimal=10)


def test_grid_with_coords():
    numPix = 11
    deltaPix = 1.0
    (
        x_grid,
        y_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    ) = util.make_grid_with_coordtransform(
        numPix, deltaPix, subgrid_res=1, left_lower=False
    )
    ra = 0
    dec = 0
    x, y = util.map_coord2pix(ra, dec, x_at_radec_0, y_at_radec_0, Mcoord2pix)
    assert x == 5
    assert y == 5

    numPix = 11
    deltaPix = 0.1
    (
        x_grid,
        y_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    ) = util.make_grid_with_coordtransform(
        numPix, deltaPix, subgrid_res=1, left_lower=False
    )
    ra = 0
    dec = 0
    x, y = util.map_coord2pix(ra, dec, x_at_radec_0, y_at_radec_0, Mcoord2pix)
    assert x == 5
    assert y == 5

    numPix = 11
    deltaPix = 1.0
    (
        x_grid,
        y_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    ) = util.make_grid_with_coordtransform(
        numPix, deltaPix, subgrid_res=1, left_lower=False, inverse=True
    )
    x_, y_ = 0, 0
    ra, dec = util.map_coord2pix(x_, y_, ra_at_xy_0, dec_at_xy_0, Mpix2coord)
    assert ra == 5
    assert dec == -5

    numPix = 11
    deltaPix = 1.0
    (
        x_grid,
        y_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    ) = util.make_grid_with_coordtransform(
        numPix, deltaPix, subgrid_res=1, left_lower=False, inverse=False
    )
    x_, y_ = 0, 0
    ra, dec = util.map_coord2pix(x_, y_, ra_at_xy_0, dec_at_xy_0, Mpix2coord)
    assert ra == -5
    assert dec == -5

    numPix = 11
    deltaPix = 0.1
    (
        x_grid,
        y_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    ) = util.make_grid_with_coordtransform(
        numPix, deltaPix, subgrid_res=1, left_lower=False
    )
    x_, y_ = 0, 0
    ra, dec = util.map_coord2pix(x_, y_, ra_at_xy_0, dec_at_xy_0, Mpix2coord)
    assert ra == 0.5
    assert dec == -0.5
    x__, y__ = util.map_coord2pix(ra, dec, x_at_radec_0, y_at_radec_0, Mcoord2pix)
    assert x__ == x_
    assert y__ == y_

    numPix = 11
    deltaPix = 0.1
    (
        x_grid,
        y_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    ) = util.make_grid_with_coordtransform(
        numPix, deltaPix, subgrid_res=1, left_lower=True
    )
    assert ra_at_xy_0 == 0
    assert dec_at_xy_0 == 0

    numPix = 11
    deltaPix = 0.1
    (
        x_grid,
        y_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    ) = util.make_grid_with_coordtransform(
        numPix, deltaPix, subgrid_res=1, left_lower=True, center_ra=2, center_dec=3
    )
    assert ra_at_xy_0 == 2
    assert dec_at_xy_0 == 3


def test_array2image():
    array = np.linspace(1, 100, 100)
    image = util.array2image(array)
    assert image[9][9] == 100
    assert image[0][9] == 10


def test_image2array():
    image = np.zeros((10, 10))
    image[1, 2] = 1
    array = util.image2array(image)
    assert array[12] == 1


def test_image2array2image():
    image = np.zeros((20, 10))
    nx, ny = np.shape(image)
    image[1, 2] = 1
    array = util.image2array(image)
    image_new = util.array2image(array, nx, ny)
    assert image_new[1, 2] == image[1, 2]


def test_array2cube():
    array = np.linspace(1, 200, 200)
    image = util.array2cube(array, 2, 100)
    assert image[0][9][9] == 100
    assert image[1][0][9] == 110


def test_cube2array():
    sube = np.zeros((2, 10, 10))
    sube[1, 2, 2] = 1
    array = util.cube2array(sube)
    assert array[122] == 1


def test_cube2array2cube():
    cube = np.zeros((2, 10, 10))
    ns, nx, ny = np.shape(cube)
    assert nx == ny  # condition required
    nxy = nx * ny
    cube[1, 2, 2] = 1
    array = util.cube2array(cube)
    cube_new = util.array2cube(array, ns, nxy)
    assert cube_new[1, 2, 2] == cube[1, 2, 2]


def test_get_axes():
    numPix = 11
    deltapix = 0.1
    x_grid, y_grid = util.make_grid(numPix, deltapix)
    x_axes, y_axes = util.get_axes(x_grid, y_grid)
    npt.assert_almost_equal(x_axes[0], -0.5, decimal=12)
    npt.assert_almost_equal(y_axes[0], -0.5, decimal=12)
    npt.assert_almost_equal(x_axes[1], -0.4, decimal=12)
    npt.assert_almost_equal(y_axes[1], -0.4, decimal=12)
    x_grid += 1
    x_axes, y_axes = util.get_axes(x_grid, y_grid)
    npt.assert_almost_equal(x_axes[0], 0.5, decimal=12)
    npt.assert_almost_equal(y_axes[0], -0.5, decimal=12)


def test_symmetry():
    array = np.linspace(0, 10, 100)
    image = util.array2image(array)
    array_new = util.image2array(image)
    assert array_new[42] == array[42]


def test_displaceAbs():
    x = np.array([0, 1, 2])
    y = np.array([3, 2, 1])
    sourcePos_x = 1
    sourcePos_y = 2
    result = util.displaceAbs(x, y, sourcePos_x, sourcePos_y)
    assert result[0] == np.sqrt(2)
    assert result[1] == 0


def test_get_distance():
    x_mins = np.array([1.0])
    y_mins = np.array([1.0])
    x_true = np.array([0.0])
    y_true = np.array([0.0])
    dist = util.get_distance(x_mins, y_mins, x_true, y_true)
    assert dist == 2

    x_mins = np.array([1.0, 2])
    y_mins = np.array([1.0, 1])
    x_true = np.array([0.0])
    y_true = np.array([0.0])
    dist = util.get_distance(x_mins, y_mins, x_true, y_true)
    assert dist == 10000000000

    x_mins = np.array([1.0, 2])
    y_mins = np.array([1.0, 1])
    x_true = np.array([0.0, 1])
    y_true = np.array([0.0, 2])
    dist = util.get_distance(x_mins, y_mins, x_true, y_true)
    assert dist == 6

    x_mins = np.array([1.0, 2, 0])
    y_mins = np.array([1.0, 1, 0])
    x_true = np.array([0.0, 1, 1])
    y_true = np.array([0.0, 2, 1])
    dist = util.get_distance(x_mins, y_mins, x_true, y_true)
    assert dist == 2


def test_selectBest():
    array = np.array([4, 3, 6, 1, 3])
    select = np.array([2, 4, 7, 3, 3])
    numSelect = 4
    array_select = util.selectBest(array, select, numSelect, highest=True)
    assert array_select[0] == 6
    assert array_select[3] == 1

    array_select = util.selectBest(array, select, numSelect, highest=False)
    assert array_select[0] == 3
    assert array_select[3] == 4

    array_select = util.selectBest(array, select, numSelect=10, highest=False)
    assert len(array_select) == len(array)


def test_select_best():
    array = np.array([4, 3, 6, 1, 3])
    select = np.array([2, 4, 7, 3, 3])
    numSelect = 4
    array_select = util.select_best(array, select, numSelect, highest=True)

    assert array_select[0] == 6
    assert array_select[3] == 1

    array_select = util.select_best(array, select, numSelect, highest=False)
    assert array_select[0] == 3
    assert array_select[3] == 4

    array_select = util.select_best(array, select, num_select=10, highest=False)
    assert len(array_select) == len(array)


def test_compare_distance():
    x_mapped = np.array([4, 3, 6, 1, 3])
    y_mapped = np.array([2, 4, 7, 3, 3])
    X2 = util.compare_distance(x_mapped, y_mapped)
    assert X2 == 140


def test_min_square_dist():
    x_1 = np.array([4, 3, 6, 1, 3])
    y_1 = np.array([2, 4, 7, 3, 3])
    x_2 = np.array([4, 3, 6, 1, 3])
    y_2 = np.array([2, 3, 7, 3, 3])
    dist = util.min_square_dist(x_1, y_1, x_2, y_2)
    assert dist[0] == 0
    assert dist[1] == 1


def test_neighbor_select_fast():
    a = np.ones(100)
    a[41] = 0
    x = np.linspace(0, 99, 100)
    y = np.linspace(0, 99, 100)
    x_mins, y_mins, values = util.local_minima_2d(a, x, y)
    assert x_mins[0] == 41
    assert y_mins[0] == 41
    assert values[0] == 0


def test_neighborSelect():
    a = np.ones(100)
    a[41] = 0
    x = np.linspace(0, 99, 100)
    y = np.linspace(0, 99, 100)
    x_mins, y_mins, values = util.neighborSelect(a, x, y)
    assert x_mins[0] == 41
    assert y_mins[0] == 41
    assert values[0] == 0


def test_make_subgrid():
    numPix = 101
    deltapix = 1
    x_grid, y_grid = util.make_grid(numPix, deltapix, subgrid_res=1)
    x_sub_grid, y_sub_grid = util.make_subgrid(x_grid, y_grid, subgrid_res=2)
    assert np.sum(x_grid) == 0
    assert len(x_grid) == 101 * 101
    assert x_sub_grid[0] == -50.25
    assert y_sub_grid[17] == -50.25

    x_sub_grid_new, y_sub_grid_new = util.make_subgrid(x_grid, y_grid, subgrid_res=4)
    assert x_sub_grid_new[0] == -50.375


def test_fwhm2sigma():
    fwhm = 0.5
    sigma = util.fwhm2sigma(fwhm)
    assert sigma == fwhm / (2 * np.sqrt(2 * np.log(2)))


def test_points_on_circle():
    radius = 1
    points = 8
    ra, dec = util.points_on_circle(radius, points, connect_ends=True)
    assert ra[0] == 1
    assert dec[0] == 0

    ra_, dec_ = util.points_on_circle(radius, points - 1, connect_ends=False)
    npt.assert_almost_equal(ra[:-1], ra_, decimal=8)
    npt.assert_almost_equal(dec[:-1], dec_, decimal=8)


def test_convert_bool_list():
    bool_list = util.convert_bool_list(n=10, k=None)
    assert len(bool_list) == 10
    assert bool_list[0] == True

    bool_list = util.convert_bool_list(n=10, k=3)
    assert len(bool_list) == 10
    assert bool_list[3] is True
    assert bool_list[2] is False

    bool_list = util.convert_bool_list(n=10, k=[3, 7])
    assert len(bool_list) == 10
    assert bool_list[3] is True
    assert bool_list[7] is True
    assert bool_list[2] is False

    bool_list = util.convert_bool_list(n=3, k=[False, False, True])
    assert len(bool_list) == 3
    assert bool_list[0] is False
    assert bool_list[1] is False
    assert bool_list[2] is True

    bool_list = util.convert_bool_list(n=3, k=[])
    assert len(bool_list) == 3
    assert bool_list[0] is False


def test_area():
    r = 1
    x_, y_ = util.points_on_circle(radius=r, connect_ends=True, num_points=1000)
    vs = np.dstack([x_, y_])[0]
    a = util.area(vs)
    npt.assert_almost_equal(a, np.pi * r**2, decimal=3)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            array = np.ones(5)
            util.array2image(array)
        with self.assertRaises(ValueError):
            array = np.ones((2, 2))
            util.array2cube(array, 2, 2)
        with self.assertRaises(ValueError):
            x, y = np.ones(6), np.ones(6)
            util.get_axes(x, y)
        with self.assertRaises(ValueError):
            util.selectBest(
                array=np.ones(6), criteria=np.ones(5), numSelect=1, highest=True
            )
        with self.assertRaises(ValueError):
            util.select_best(
                array=np.ones(6), criteria=np.ones(5), num_select=1, highest=True
            )
        with self.assertRaises(ValueError):
            util.convert_bool_list(n=2, k=[3, 7])
        with self.assertRaises(ValueError):
            util.convert_bool_list(n=3, k=[True, True])
        with self.assertRaises(ValueError):
            util.convert_bool_list(n=2, k=[0.1, True])

    def test_raise_make_grid(self):
        with self.assertRaises(ValueError):
            util.make_grid(numPix=1.1, deltapix=1)
        with self.assertRaises(ValueError):
            util.make_grid(numPix=[1.1, 1], deltapix=1)


if __name__ == "__main__":
    pytest.main()



import numpy as np
import numpy.testing as npt
import lenstronomy.SimulationAPI.observation_constructor as constructor
import pytest
import lenstronomy.Util.data_util as data_util


class TestPointSourceVariability(object):
    def setup_method(self):
        pass

    def test_image(self):
        # we define a time variable function in magnitude space
        def var_func(time):
            sigma = 100
            mag_0 = 30
            cps = np.exp(-(time**2) / (2 * sigma**2))
            mag = data_util.cps2magnitude(cps, magnitude_zero_point=0)
            mag_norm = data_util.cps2magnitude(1, magnitude_zero_point=0)
            mag_return = -mag + mag_norm + mag_0
            return mag_return

        kwargs_model_time_var = {
            "lens_model_list": ["SPEP", "SHEAR"],  # list of lens models to be used
            "lens_light_model_list": ["SERSIC_ELLIPSE"],
            # list of unlensed light models to be used
            "source_light_model_list": ["SERSIC_ELLIPSE"],
            # list of extended source models to be used
            "z_lens": 0.5,
            "z_source": 2,
        }
        instrument_name = "LSST"
        observation_name = "LSST_g_band"
        kwargs_single_band = constructor.observation_constructor(
            instrument_name=instrument_name, observation_name=observation_name
        )

        kwargs_single_band["data_count_unit"] = "e-"
        kwargs_numerics = {}
        numpix = 20
        # source position
        source_x, source_y = 0.01, 0.1
        # lens light
        kwargs_lens_light_mag_g = [
            {
                "magnitude": 100,
                "R_sersic": 0.6,
                "n_sersic": 4,
                "e1": 0.1,
                "e2": -0.1,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        # source light
        kwargs_source_mag_g = [
            {
                "magnitude": 100,
                "R_sersic": 0.3,
                "n_sersic": 1,
                "e1": -0.3,
                "e2": -0.2,
                "center_x": 0,
                "center_y": 0,
            }
        ]

        kwargs_lens = [
            {
                "theta_E": 1,
                "gamma": 2,
                "e1": 0.1,
                "e2": -0.1,
                "center_x": 0,
                "center_y": 0,
            },  # SIE model
            {"gamma1": 0.03, "gamma2": 0.01},  # SHEAR model
        ]

        from lenstronomy.SimulationAPI.point_source_variability import (
            PointSourceVariability,
        )

        ps_var = PointSourceVariability(
            source_x,
            source_y,
            var_func,
            numpix,
            kwargs_single_band,
            kwargs_model_time_var,
            kwargs_numerics,
            kwargs_lens,
            kwargs_source_mag_g,
            kwargs_lens_light_mag_g,
            kwargs_ps_mag=None,
        )

        time = 0
        image_g = ps_var.image_time(time=time)
        npt.assert_almost_equal(np.sum(image_g), 8, decimal=1)

        t_days = ps_var.delays
        assert len(t_days) == 4


if __name__ == "__main__":
    pytest.main()



from lenstronomy.GalKin import velocity_util
from lenstronomy.Util import kernel_util


from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class PSF(object):
    """General class to handle the PSF in the GalKin module for rendering the
    displacement of photons/spectro."""

    def __init__(self, psf_type, **kwargs_psf):
        """

        :param psf_type: string, point spread function type, current support for 'GAUSSIAN' and 'MOFFAT'
        :param kwargs_psf: keyword argument describing the relevant parameters of the PSF.
        """
        if psf_type == "GAUSSIAN":
            self._psf = PSFGaussian(**kwargs_psf)
        elif psf_type == "MOFFAT":
            self._psf = PSFMoffat(**kwargs_psf)
        else:
            raise ValueError("psf_type %s not supported for convolution!" % psf_type)

    def displace_psf(self, x, y):
        """

        :param x: x-coordinate of light ray
        :param y: y-coordinate of light ray
        :return: x', y' displaced by the two-dimensional PSF distribution function
        """
        return self._psf.displace_psf(x, y)

    def convolution_kernel(self, delta_pix, num_pix=21):
        """Normalized convolution kernel.

        :param delta_pix: pixel scale of kernel
        :param num_pix: number of pixels per axis of the kernel
        :return: 2d numpy array of kernel
        """
        return self._psf.convolution_kernel(delta_pix, num_pix)


@export
class PSFGaussian(object):
    """Gaussian PSF."""

    def __init__(self, fwhm):
        """

        :param fwhm: full width at half maximum seeing condition
        """
        self._fwhm = fwhm

    def displace_psf(self, x, y):
        """

        :param x: x-coordinate of light ray
        :param y: y-coordinate of light ray
        :return: x', y' displaced by the two-dimensional PSF distribution function
        """
        return velocity_util.displace_PSF_gaussian(x, y, self._fwhm)

    def convolution_kernel(self, delta_pix, num_pix=21):
        """Normalized convolution kernel.

        :param delta_pix: pixel scale of kernel
        :param num_pix: number of pixels per axis of the kernel
        :return: 2d numpy array of kernel
        """

        kernel = kernel_util.kernel_gaussian(num_pix, delta_pix, self._fwhm)
        return kernel


@export
class PSFMoffat(object):
    """Moffat PSF."""

    def __init__(self, fwhm, moffat_beta):
        """

        :param fwhm: full width at half maximum seeing condition
        :param moffat_beta: float, beta parameter of Moffat profile
        """
        self._fwhm = fwhm
        self._moffat_beta = moffat_beta

    def displace_psf(self, x, y):
        """

        :param x: x-coordinate of light ray
        :param y: y-coordinate of light ray
        :return: x', y' displaced by the two-dimensional PSF distribution function
        """
        return velocity_util.displace_PSF_moffat(x, y, self._fwhm, self._moffat_beta)

    def convolution_kernel(self, delta_pix, num_pix=21):
        """Normalized convolution kernel.

        :param delta_pix: pixel scale of kernel
        :param num_pix: number of pixels per axis of the kernel
        :return: 2d numpy array of kernel
        """

        kernel = kernel_util.kernel_moffat(
            num_pix=num_pix,
            delta_pix=delta_pix,
            fwhm=self._fwhm,
            moffat_beta=self._moffat_beta,
        )
        return kernel



from lenstronomy.Sampling.Likelihoods.kinematic_NN_call import KinematicNN
import numpy as np
import numpy.testing as npt
import copy
import matplotlib.pyplot as plt


class TestKinNNCall(object):
    def setup(self):
        self.example_input = np.array(
            [
                8.56306356e-01,
                6.77923305e-01,
                1.46690845e00,
                3.00871520e00,
                1.30421229e00,
                8.0e-2,
                4.55349851e-01,
                -5.68615913e-02,
                8.07435011e01,
            ]
        )
        self.kinematic_NN = KinematicNN()

    def test_generate_map(self):
        if self.kinematic_NN.SKiNN_installed:
            map = self.kinematic_NN.generate_map(input_p=self.example_input)
            # check that returned map is 551x551 pixels at approximately 200 km/s
            npt.assert_equal(np.shape(map), (551, 551))
            npt.assert_allclose(np.mean(map), 235.19, rtol=1e-2)
            # plt.imshow(map)
            # plt.colorbar()
            # plt.show()

    def test_check_bounds(self):
        if self.kinematic_NN.SKiNN_installed:
            verbose = True
            in_bounds = self.kinematic_NN.check_bounds(
                self.example_input, same_orientation=True, verbose=verbose
            )
            assert in_bounds is True
            # check if different orientations
            in_bounds = self.kinematic_NN.check_bounds(
                self.example_input, same_orientation=False, verbose=verbose
            )
            assert in_bounds is False
            # check all params out of bounds
            for i in range(len(self.example_input)):
                test_input = copy.copy(self.example_input)
                test_input[i] = 100
                in_bounds = self.kinematic_NN.check_bounds(
                    test_input, same_orientation=True, verbose=verbose
                )
                assert in_bounds is False

            # check r_eff > theta_E
            test_input = copy.copy(self.example_input)
            test_input[4] = 1.6
            in_bounds = self.kinematic_NN.check_bounds(
                test_input, same_orientation=True, verbose=verbose
            )
            assert in_bounds is False

            # check inclination
            test_input = copy.copy(self.example_input)
            test_input[8] = 10
            in_bounds = self.kinematic_NN.check_bounds(
                test_input, same_orientation=True, verbose=verbose
            )
            assert in_bounds is False

    def test_plot_map(self):
        if self.kinematic_NN.SKiNN_installed:
            f, ax = plt.subplots(1, 1, figsize=(4, 4))
            ax = self.kinematic_NN.plot_map(ax=ax, input_p=self.example_input)
            plt.close()



__author__ = "sibirrer"

from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
import numpy as np

__all__ = ["Shift"]


class Shift(LensProfileBase):
    """Lens model with a constant shift of the deflection field."""

    param_names = ["alpha_x", "alpha_y"]
    lower_limit_default = {"alpha_x": -1000, "alpha_y": -1000}
    upper_limit_default = {"alpha_x": 1000, "alpha_y": 1000}

    def function(self, x, y, alpha_x, alpha_y):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param alpha_x: shift in x-direction (angle)
        :param alpha_y: shift in y-direction (angle)
        :return: lensing potential
        """

        return np.zeros_like(x)

    def derivatives(self, x, y, alpha_x, alpha_y):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param alpha_x: shift in x-direction (angle)
        :param alpha_y: shift in y-direction (angle)
        :return: deflection in x- and y-direction
        """
        f_x = np.ones_like(x) * alpha_x
        f_y = np.ones_like(x) * alpha_y
        return f_x, f_y

    def hessian(self, x, y, alpha_x, alpha_y):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param alpha_x: shift in x-direction (angle)
        :param alpha_y: shift in y-direction (angle)
        :return: hessian elements f_xx, f_xy, f_yx, f_yy
        """
        f_xx, f_xy, f_yx, f_yy = 0, 0, 0, 0
        return f_xx, f_xy, f_yx, f_yy



from lenstronomy.LensModel.Profiles.p_jaffe import PJaffe
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
import numpy as np

__all__ = ["PJaffe_Ellipse"]


class PJaffe_Ellipse(LensProfileBase):
    """
    class to compute the DUAL PSEUDO ISOTHERMAL ELLIPTICAL MASS DISTRIBUTION
    based on Eliasdottir (2007) https://arxiv.org/pdf/0710.5636.pdf Appendix A
    with the ellipticity implemented in the potential

    Module name: 'PJAFFE_ELLIPSE';

    An alternative name is dPIED.

    The 3D density distribution is

    .. math::
        \\rho(r) = \\frac{\\rho_0}{(1+r^2/Ra^2)(1+r^2/Rs^2)}

    with :math:`Rs > Ra`.

    The projected density is

    .. math::
        \\Sigma(R) = \\Sigma_0 \\frac{Ra Rs}{Rs-Ra}\\left(\\frac{1}{\\sqrt{Ra^2+R^2}} - \\frac{1}{\\sqrt{Rs^2+R^2}} \\right)

    with

    .. math::
        \\Sigma_0 = \\pi \\rho_0 \\frac{Ra Rs}{Rs + Ra}

    In the lensing parameterization,

    .. math::
        \\sigma_0 = \\frac{\\Sigma_0}{\\Sigma_{\\rm crit}}

    """

    param_names = ["sigma0", "Ra", "Rs", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "sigma0": 0,
        "Ra": 0,
        "Rs": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "sigma0": 10,
        "Ra": 100,
        "Rs": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.spherical = PJaffe()
        self._diff = 0.000001
        super(PJaffe_Ellipse, self).__init__()

    def function(self, x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0):
        """Returns double integral of NFW profile."""
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        f_ = self.spherical.function(x_, y_, sigma0, Ra, Rs)
        return f_

    def derivatives(self, x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function (integral of NFW)"""
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        e = param_util.q2e(q)
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)
        f_x_prim, f_y_prim = self.spherical.derivatives(
            x_, y_, sigma0, Ra, Rs, center_x=0, center_y=0
        )
        f_x_prim *= np.sqrt(1 - e)
        f_y_prim *= np.sqrt(1 + e)
        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        alpha_ra, alpha_dec = self.derivatives(
            x, y, sigma0, Ra, Rs, e1, e2, center_x, center_y
        )
        diff = self._diff
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff, y, sigma0, Ra, Rs, e1, e2, center_x, center_y
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff, sigma0, Ra, Rs, e1, e2, center_x, center_y
        )

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        f_yx = (alpha_dec_dx - alpha_dec) / diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff

        return f_xx, f_xy, f_yx, f_yy

    def mass_3d_lens(self, r, sigma0, Ra, Rs, e1=0, e2=0):
        """

        :param r:
        :param sigma0:
        :param Ra:
        :param Rs:
        :param e1:
        :param e2:
        :return:
        """
        return self.spherical.mass_3d_lens(r, sigma0, Ra, Rs)



from lenstronomy.Util import util
from lenstronomy.LensModel import convergence_integrals
import numpy as np
from lenstronomy.Util import constants as const
from lenstronomy.Cosmo.lens_cosmo import LensCosmo
from lenstronomy.LensModel.lens_model import LensModel

__all__ = ["LightCone", "MassSlice"]


class LightCone(object):
    """Class to perform multi-plane ray-tracing from convergence maps at different
    redshifts From the convergence maps the deflection angles and lensing potential are
    computed (from different settings) and then an interpolated grid of all those
    quantities generate an instance of the lenstronomy LensModel multi-plane instance.
    All features of the LensModel module are supported.

    Improvements that can be made for accuracy and speed:
    1. adaptive mesh integral for the convergence map
    2. Interpolated deflection map on different scales than the mass map.

    The design principles should allow those implementations 'under the hook' of this class.
    """

    def __init__(self, mass_map_list, grid_spacing_list, redshift_list):
        """

        :param mass_map_list: 2d numpy array of mass map
         (in units physical Solar masses enclosed in each pixel/gird point of the map)
        :param grid_spacing_list: list of grid spacing of the individual mass maps
         in units of physical Mpc
        :param redshift_list: list of redshifts of the mass maps

        """
        self._mass_slice_list = []
        for i in range(len(mass_map_list)):
            self._mass_slice_list.append(
                MassSlice(mass_map_list[i], grid_spacing_list[i], redshift_list[i])
            )
        self._mass_map_list = mass_map_list
        self._grid_spacing_list = grid_spacing_list
        self._redshift_list = redshift_list

    def cone_instance(self, z_source, cosmo, multi_plane=True, kwargs_interp=None):
        """

        :param z_source: redshift to where lensing quantities are computed
        :param cosmo: astropy.cosmology class
        :param multi_plane: boolean, if True, computes multi-plane ray-tracing
        :param kwargs_interp: interpolation keyword arguments specifying the numerics.
         See description in the Interpolate() class. Only applicable for 'INTERPOL' and 'INTERPOL_SCALED' models.
        :return: LensModel instance, keyword argument list of lens model
        """
        lens_model = LensModel(
            lens_model_list=["INTERPOL"] * len(self._mass_map_list),
            lens_redshift_list=self._redshift_list,
            multi_plane=multi_plane,
            z_source_convention=z_source,
            cosmo=cosmo,
            z_source=z_source,
            kwargs_interp=kwargs_interp,
        )
        kwargs_lens = []
        for mass_slice in self._mass_slice_list:
            kwargs_lens.append(mass_slice.interpol_instance(z_source, cosmo))
        return lens_model, kwargs_lens


class MassSlice(object):
    """Class to describe a single mass slice."""

    def __init__(self, mass_map, grid_spacing, redshift):
        """

        :param mass_map: 2d numpy array of mass map (in units physical Msol)
        :param grid_spacing: grid spacing of the mass map (in units physical Mpc)
        :param redshift: redshift
        """
        nx, ny = np.shape(mass_map)
        if nx != ny:
            raise ValueError(
                "Shape of mass map needs to be square!, set as %s %s" % (nx, ny)
            )
        self._mass_map = mass_map
        self._grid_spacing = grid_spacing
        self._redshift = redshift
        (
            self._f_x_mass,
            self._f_y_mass,
        ) = convergence_integrals.deflection_from_kappa_grid(
            self._mass_map, self._grid_spacing
        )
        self._f_mass = convergence_integrals.potential_from_kappa_grid(
            self._mass_map, self._grid_spacing
        )
        x_grid, y_grid = util.make_grid(
            numPix=len(self._mass_map), deltapix=self._grid_spacing
        )
        self._x_axes_mpc, self._y_axes_mpc = util.get_axes(x_grid, y_grid)

    def interpol_instance(self, z_source, cosmo):
        """Scales the mass map integrals (with units of mass not convergence) into a
        convergence map for the given cosmology and source redshift and returns the
        keyword arguments of the interpolated reduced deflection and lensing potential.

        :param z_source: redshift of the source
        :param cosmo: astropy.cosmology instance
        :return: keyword arguments of the interpolation instance with numerically
            computed deflection angles and lensing potential
        """
        lens_cosmo = LensCosmo(z_lens=self._redshift, z_source=z_source, cosmo=cosmo)
        mpc2arcsec = lens_cosmo.dd * const.arcsec
        x_axes = self._x_axes_mpc / mpc2arcsec  # units of arc seconds in grid spacing
        y_axes = self._y_axes_mpc / mpc2arcsec  # units of arc seconds in grid spacing

        f_ = self._f_mass / lens_cosmo.sigma_crit_angle / self._grid_spacing**2
        f_x = (
            self._f_x_mass
            / lens_cosmo.sigma_crit_angle
            / self._grid_spacing**2
            * mpc2arcsec
        )
        f_y = (
            self._f_y_mass
            / lens_cosmo.sigma_crit_angle
            / self._grid_spacing**2
            * mpc2arcsec
        )
        kwargs_interp = {
            "grid_interp_x": x_axes,
            "grid_interp_y": y_axes,
            "f_": f_,
            "f_x": f_x,
            "f_y": f_y,
        }
        return kwargs_interp



from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Plots.plot_quasar_images import plot_quasar_images
import matplotlib.pyplot as plt
import pytest


class TestPlotQuasarImages(object):
    def test_plot_quasar_images(self):
        lens_model_list = ["EPL", "SHEAR"]
        z_source = 1.5
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": 2.0,
                "e1": 0.02,
                "e2": -0.09,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.01, "gamma2": 0.03},
        ]
        lensmodel = LensModel(lens_model_list)
        solver = LensEquationSolver(lensmodel)
        source_x, source_y = 0.07, 0.03
        x_image, y_image = solver.findBrightImage(source_x, source_y, kwargs_lens)
        source_fwhm_parsec = 40.0

        plot_quasar_images(
            lensmodel,
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
        )
        plt.close()


if __name__ == "__main__":
    pytest.main()



__author__ = "ewoudwempe"

import numpy as np
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.Util.numba_util import jit, nan_to_num

__all__ = ["EPL_numba"]


class EPL_numba(LensProfileBase):
    """ "
    Elliptical Power Law mass profile - computation accelerated with numba

    .. math::
        \\kappa(x, y) = \\frac{3-\\gamma}{2} \\left(\\frac{\\theta_{E}}{\\sqrt{q x^2 + y^2/q}} \\right)^{\\gamma-1}

    with :math:`\\theta_{E}` is the (circularized) Einstein radius,
    :math:`\\gamma` is the negative power-law slope of the 3D mass distributions,
    :math:`q` is the minor/major axis ratio,
    and :math:`x` and :math:`y` are defined in a coordinate system aligned with the major and minor axis of the lens.

    In terms of eccentricities, this profile is defined as

    .. math::
        \\kappa(r) = \\frac{3-\\gamma}{2} \\left(\\frac{\\theta'_{E}}{r \\sqrt{1 − e*\\cos(2*\\phi)}} \\right)^{\\gamma-1}

    with :math:`\\epsilon` is the ellipticity defined as

    .. math::
        \\epsilon = \\frac{1-q^2}{1+q^2}

    And an Einstein radius :math:`\\theta'_{\\rm E}` related to the definition used is

    .. math::
        \\left(\\frac{\\theta'_{\\rm E}}{\\theta_{\\rm E}}\\right)^{2} = \\frac{2q}{1+q^2}.


    The mathematical form of the calculation is presented by Tessore & Metcalf (2015), https://arxiv.org/abs/1507.01819.
    The current implementation is using hyperbolic functions. The paper presents an iterative calculation scheme,
    converging in few iterations to high precision and accuracy.

    A (slower) implementation of the same model using hyperbolic functions without the iterative calculation
    is accessible as 'EPL' not requiring numba.
    """

    param_names = ["theta_E", "gamma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 1.5,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 2.5,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        super(EPL_numba).__init__()

    @staticmethod
    @jit()
    def function(x, y, theta_E, gamma, e1, e2, center_x=0.0, center_y=0.0):
        """

        :param x: x-coordinate (angle)
        :param y: y-coordinate (angle)
        :param theta_E: Einstein radius (angle), pay attention to specific definition!
        :param gamma: logarithmic slope of the power-law profile. gamma=2 corresponds to isothermal
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: x-position of lens center
        :param center_y: y-position of lens center
        :return: lensing potential
        """
        z, b, t, q, ang = param_transform(
            x, y, theta_E, gamma, e1, e2, center_x, center_y
        )
        alph = alpha(z.real, z.imag, b, q, t)
        return 1 / (2 - t) * (z.real * alph.real + z.imag * alph.imag)

    @staticmethod
    @jit()
    def derivatives(x, y, theta_E, gamma, e1, e2, center_x=0.0, center_y=0.0):
        """

        :param x: x-coordinate (angle)
        :param y: y-coordinate (angle)
        :param theta_E: Einstein radius (angle), pay attention to specific definition!
        :param gamma: logarithmic slope of the power-law profile. gamma=2 corresponds to isothermal
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: x-position of lens center
        :param center_y: y-position of lens center
        :return: deflection angles alpha_x, alpha_y
        """
        z, b, t, q, ang = param_transform(
            x, y, theta_E, gamma, e1, e2, center_x, center_y
        )
        alph = alpha(z.real, z.imag, b, q, t) * np.exp(1j * ang)
        return alph.real, alph.imag

    @staticmethod
    @jit()
    def hessian(x, y, theta_E, gamma, e1, e2, center_x=0.0, center_y=0.0):
        """

        :param x: x-coordinate (angle)
        :param y: y-coordinate (angle)
        :param theta_E: Einstein radius (angle), pay attention to specific definition!
        :param gamma: logarithmic slope of the power-law profile. gamma=2 corresponds to isothermal
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: x-position of lens center
        :param center_y: y-position of lens center
        :return: Hessian components f_xx, f_yy, f_xy
        """
        z, b, t, q, ang_ell = param_transform(
            x, y, theta_E, gamma, e1, e2, center_x, center_y
        )
        ang = np.angle(z)
        # r = np.abs(z)
        zz_ell = z.real * q + 1j * z.imag
        R = np.abs(zz_ell)
        phi = np.angle(zz_ell)

        # u = np.minimum(nan_to_num((b/R)**t),1e100)
        u = np.fmin(
            (b / R) ** t, 1e10
        )  # I remove all factors of (b/R)**t to only have to remove nans once.
        # The np.fmin is a regularisation near R=0, to avoid overflows
        # in the magnification calculations
        kappa = (2 - t) / 2
        Roverr = np.sqrt(np.cos(ang) ** 2 * q**2 + np.sin(ang) ** 2)

        Omega = omega(phi, t, q)
        alph = (2 * b) / (1 + q) / b * Omega
        gamma_shear = (
            -np.exp(2j * (ang + ang_ell)) * kappa
            + (1 - t) * np.exp(1j * (ang + 2 * ang_ell)) * alph * Roverr
        )

        f_xx = (kappa + gamma_shear.real) * u
        f_yy = (kappa - gamma_shear.real) * u
        f_xy = gamma_shear.imag * u
        # Fix the nans if x=y=0 is filled in

        return f_xx, f_xy, f_xy, f_yy


@jit()
def param_transform(x, y, theta_E, gamma, e1, e2, center_x=0.0, center_y=0.0):
    """Converts the parameters from lenstronomy definitions (as defined in PEMD) to the
    definitions of Tessore+ (2015)"""
    t = gamma - 1
    phi_G, q = param_util.ellipticity2phi_q(e1, e2)
    x_shift = x - center_x
    y_shift = y - center_y
    ang = phi_G
    z = np.exp(-1j * phi_G) * (x_shift + y_shift * 1j)
    return z, theta_E * np.sqrt(q), t, q, ang


@jit()
def alpha(x, y, b, q, t, Omega=None):
    """Calculates the complex deflection.

    :param x: x-coordinate (angle)
    :param y: y-coordinate (angle)
    :param b: Einstein radius (angle), pay attention to specific definition!
    :param q: axis ratio
    :param t: logarithmic power-law slope. Is t=gamma-1
    :param Omega: If given, use this Omega (to avoid recalculations)
    :return: complex deflection angle
    """
    zz = x * q + 1j * y
    R = np.abs(zz)
    phi = np.angle(zz)
    if Omega is None:
        Omega = omega(phi, t, q)
    # Omega = omega(phi, t, q)
    alph = (2 * b) / (1 + q) * nan_to_num((b / R) ** t * R / b) * Omega
    return alph


@jit(
    fastmath=True
)  # Because of the reduction nature of this, relaxing commutativity actually matters a lot (4x speedup).
def omega(phi, t, q, niter_max=200, tol=1e-16):
    f = (1 - q) / (1 + q)
    omegas = np.zeros_like(phi, dtype=np.complex128)
    niter = min(
        niter_max, int(np.log(tol) / np.log(f)) + 2
    )  # The absolute value of each summand is always less than f, hence this limit for the number of iterations.
    Omega = 1 * np.exp(1j * phi)
    fact = -f * np.exp(2j * phi)
    for n in range(1, niter):
        omegas += Omega
        Omega *= (2 * n - (2 - t)) / (2 * n + (2 - t)) * fact
    omegas += Omega
    return omegas



__author__ = "sibirrer"

import numpy as np
import numpy.polynomial.hermite as hermite
import math

import lenstronomy.Util.util as util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Shapelets(object):
    """Class for 2d cartesian Shapelets.

    Sources:
    Refregier 2003: Shapelets: I. A Method for Image Analysis https://arxiv.org/abs/astro-ph/0105178
    Refregier 2003: Shapelets: II. A Method for Weak Lensing Measurements https://arxiv.org/abs/astro-ph/0105179

    For one dimension, the shapelets are defined as

    .. math::
        \\phi_n(x) \\equiv \\left[2^n \\pi^{1/2} n!  \\right]]^{-1/2}H_n(x) e^{-\\frac{x^2}{2}}

    This basis is orthonormal. The dimensional basis function is

    .. math::
        B_n(x;\\beta) \\equiv \\beta^{-1/2} \\phi_n(\\beta^{-1}x)

    which are orthonormal as well.

    The two-dimensional basis function is

    .. math::
        \\phi_{\\bf n}({\bf x}) \\equiv \\phi_{n1}(x1) \\phi_{n2}(x2)

    where :math:`{\\bf n} \\equiv (n1, n2)` and :math:`{\\bf x} \\equiv (x1, x2)`.

    The dimensional two-dimentional basis function is

    .. math::
        B_{\\bf n}({\\bf x};\\beta) \\equiv \\beta^{-1/2} \\phi_{\\bf n}(\\beta^{-1}{\\bf x}).
    """

    param_names = ["amp", "beta", "n1", "n2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "beta": 0.01,
        "n1": 0,
        "n2": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "beta": 100,
        "n1": 150,
        "n2": 150,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(
        self, interpolation=False, precalc=False, stable_cut=True, cut_scale=5
    ):
        """Load interpolation of the Hermite polynomials in a range [-30,30] in order
        n<= 150.

        :param interpolation: boolean; if True, uses interpolated pre-calculated
            shapelets in the evaluation
        :param precalc: boolean; if True interprets as input (x, y) as pre-calculated
            normalized shapelets
        :param stable_cut: boolean; if True, sets the values outside of
            :math:`\\sqrt\\left(n_{\\rm max} + 1 \\right) \\beta s_{\\rm cut scale} =
            0`.
        :param cut_scale: float, scaling parameter where to cut the shapelets. This is
            for numerical reasons such that the polynomials in the Hermite function do
            not get unstable.
        """

        self._interpolation = interpolation
        self._precalc = precalc
        self._stable_cut = stable_cut
        self._cut_scale = cut_scale
        if interpolation:
            n_order = 50
            self.H_interp = [[] for _ in range(0, n_order)]
            self.x_grid = np.linspace(-50, 50, 6000)
            for k in range(0, n_order):
                n_array = np.zeros(k + 1)
                n_array[k] = 1
                values = self.hermval(self.x_grid, n_array)
                self.H_interp[k] = values

    def hermval(self, x, n_array, tensor=True):
        """
        computes the Hermit polynomial as numpy.polynomial.hermite.hermval
        difference: for values more than sqrt(n_max + 1) * cut_scale, the value is set to zero
        this should be faster and numerically stable

        :param x: array of values
        :param n_array: list of coeffs in H_n
        :param tensor: see numpy.polynomial.hermite.hermval
        :return: see numpy.polynomial.hermite.hermval
        """
        if not self._stable_cut:
            return hermite.hermval(x, n_array, tensor=tensor)
        else:
            n_max = len(n_array)
            x_cut = np.sqrt(n_max + 1) * self._cut_scale
            if isinstance(x, int) or isinstance(x, float):
                if x >= x_cut:
                    return 0
                else:
                    return hermite.hermval(x, n_array)
            else:
                out = np.zeros_like(x)
                out[x < x_cut] = hermite.hermval(x[x < x_cut], n_array, tensor=tensor)
                return out

    def function(self, x, y, amp, beta, n1, n2, center_x, center_y):
        """2d cartesian shapelet.

        :param x: x-coordinate
        :param y: y-coordinate
        :param amp: amplitude of shapelet
        :param beta: scale factor of shapelet
        :param n1: x-order of Hermite polynomial
        :param n2: y-order of Hermite polynomial
        :param center_x: center in x
        :param center_y: center in y
        :return: flux surface brightness at (x, y)
        """

        if self._precalc:
            return amp * x[n1] * y[n2]  # / beta
        x_ = x - center_x
        y_ = y - center_y
        return np.nan_to_num(
            amp * self.phi_n(n1, x_ / beta) * self.phi_n(n2, y_ / beta)
        )  # /beta

    def H_n(self, n, x):
        """Constructs the Hermite polynomial of order n at position x (dimensionless)

        :param n: The n'the basis function.
        :param x: 1-dim position (dimensionless)
        :type x: float or numpy array.
        :returns: array-- H_n(x).
        """
        if not self._interpolation:
            n_array = np.zeros(n + 1)
            n_array[n] = 1
            return self.hermval(
                x, n_array, tensor=False
            )  # attention, this routine calculates every single hermite polynomial and multiplies it with zero (exept the right one)
        else:
            return np.interp(x, self.x_grid, self.H_interp[n])

    def phi_n(self, n, x):
        """Constructs the 1-dim basis function (formula (1) in Refregier et al. 2001)

        :param n: The n'the basis function.
        :type n: int.
        :param x: 1-dim position (dimensionless)
        :type x: float or numpy array.
        :returns: array-- phi_n(x).
        """
        prefactor = 1.0 / np.sqrt(2**n * np.sqrt(np.pi) * math.factorial(n))
        return prefactor * self.H_n(n, x) * np.exp(-(x**2) / 2.0)

    def pre_calc(self, x, y, beta, n_order, center_x, center_y):
        """Calculates the H_n(x) and H_n(y) for a given x-array and y-array for the full
        order in the polynomials.

        :param x: x-coordinates (numpy array)
        :param y: 7-coordinates (numpy array)
        :param beta: shapelet scale
        :param n_order: order of shapelets
        :param center_x: shapelet center
        :param center_y: shapelet center
        :return: list of H_n(x) and H_n(y)
        """
        x_ = x - center_x
        y_ = y - center_y
        n = len(np.atleast_1d(x))
        H_x = np.empty((n_order + 1, n))
        H_y = np.empty((n_order + 1, n))
        exp_x = np.exp(-((x_ / beta) ** 2) / 2.0)
        exp_y = np.exp(-((y_ / beta) ** 2) / 2.0)
        if n_order > 170:
            raise ValueError("polynomial order to large", n_order)
        for n in range(0, n_order + 1):
            prefactor = 1.0 / np.sqrt(2**n * np.sqrt(np.pi) * math.factorial(n))
            n_array = np.zeros(n + 1)
            n_array[n] = 1
            H_x[n] = self.hermval(x_ / beta, n_array, tensor=False) * prefactor * exp_x
            H_y[n] = self.hermval(y_ / beta, n_array, tensor=False) * prefactor * exp_y
        return H_x, H_y


@export
class ShapeletSet(object):
    """Class to operate on entire shapelet set limited by a maximal polynomial order
    n_max, such that n1 + n2 <= n_max."""

    param_names = ["amp", "n_max", "beta", "center_x", "center_y"]
    lower_limit_default = {"beta": 0.01, "center_x": -100, "center_y": -100}
    upper_limit_default = {"beta": 100, "center_x": 100, "center_y": 100}

    def __init__(self):
        self.shapelets = Shapelets(precalc=True)

    def function(self, x, y, amp, n_max, beta, center_x=0, center_y=0):
        """

        :param x: x-coordinates
        :param y: y-coordinates
        :param amp: array of amplitudes in pre-defined order of shapelet basis functions
        :param beta: shapelet scale
        :param n_max: maximum polynomial order in Hermite polynomial
        :param center_x: shapelet center
        :param center_y: shapelet center
        :return: surface brightness of combined shapelet set
        """
        num_param = int((n_max + 1) * (n_max + 2) / 2)
        f_ = np.zeros(len(np.atleast_1d(x)))
        n1 = 0
        n2 = 0
        H_x, H_y = self.shapelets.pre_calc(x, y, beta, n_max, center_x, center_y)
        for i in range(num_param):
            kwargs_source_shapelet = {
                "center_x": center_x,
                "center_y": center_y,
                "n1": n1,
                "n2": n2,
                "beta": beta,
                "amp": amp[i],
            }
            out = self.shapelets.function(H_x, H_y, **kwargs_source_shapelet)
            f_ += out
            if n1 == 0:
                n1 = n2 + 1
                n2 = 0
            else:
                n1 -= 1
                n2 += 1
        try:
            len(x)
        except:
            f_ = f_[0]
        return np.nan_to_num(f_)

    def function_split(self, x, y, amp, n_max, beta, center_x=0, center_y=0):
        """Splits shapelet set in list of individual shapelet basis function responses.

        :param x: x-coordinates
        :param y: y-coordinates
        :param amp: array of amplitudes in pre-defined order of shapelet basis functions
        :param beta: shapelet scale
        :param n_max: maximum polynomial order in Hermite polynomial
        :param center_x: shapelet center
        :param center_y: shapelet center
        :return: list of individual shapelet basis function responses
        """
        num_param = int((n_max + 1) * (n_max + 2) / 2)
        A = []
        n1 = 0
        n2 = 0
        H_x, H_y = self.shapelets.pre_calc(x, y, beta, n_max, center_x, center_y)
        for i in range(num_param):
            kwargs_source_shapelet = {
                "center_x": center_x,
                "center_y": center_y,
                "n1": n1,
                "n2": n2,
                "beta": beta,
                "amp": amp[i],
            }
            A.append(self.shapelets.function(H_x, H_y, **kwargs_source_shapelet))
            if n1 == 0:
                n1 = n2 + 1
                n2 = 0
            else:
                n1 -= 1
                n2 += 1
        return A

    def shapelet_basis_2d(
        self, num_order, beta, numPix, deltaPix=1, center_x=0, center_y=0
    ):
        """

        :param num_order: max shapelet order
        :param beta: shapelet scale
        :param numPix: number of pixel of the grid
        :return: list of shapelets drawn on pixel grid, centered.
        """
        num_param = int((num_order + 2) * (num_order + 1) / 2)
        kernel_list = []
        x_grid, y_grid = util.make_grid(numPix, deltapix=deltaPix, subgrid_res=1)
        n1 = 0
        n2 = 0
        H_x, H_y = self.shapelets.pre_calc(
            x_grid, y_grid, beta, num_order, center_x=center_x, center_y=center_y
        )
        for i in range(num_param):
            kwargs_source_shapelet = {
                "center_x": 0,
                "center_y": 0,
                "n1": n1,
                "n2": n2,
                "beta": beta,
                "amp": 1,
            }
            kernel = self.shapelets.function(H_x, H_y, **kwargs_source_shapelet)
            kernel = util.array2image(kernel)
            kernel_list.append(kernel)
            if n1 == 0:
                n1 = n2 + 1
                n2 = 0
            else:
                n1 -= 1
                n2 += 1
        return kernel_list

    def decomposition(self, image, x, y, n_max, beta, deltaPix, center_x=0, center_y=0):
        """Decomposes an image into the shapelet coefficients in same order as for the
        function call.

        :param image:
        :param x:
        :param y:
        :param n_max:
        :param beta:
        :param center_x:
        :param center_y:
        :return:
        """
        num_param = int((n_max + 1) * (n_max + 2) / 2)
        param_list = np.zeros(num_param)
        amp_norm = 1.0 / beta**2 * deltaPix**2
        n1 = 0
        n2 = 0
        H_x, H_y = self.shapelets.pre_calc(x, y, beta, n_max, center_x, center_y)
        for i in range(num_param):
            kwargs_source_shapelet = {
                "center_x": center_x,
                "center_y": center_y,
                "n1": n1,
                "n2": n2,
                "beta": beta,
                "amp": amp_norm,
            }
            base = self.shapelets.function(H_x, H_y, **kwargs_source_shapelet)
            param = np.sum(image * base)
            param_list[i] = param
            if n1 == 0:
                n1 = n2 + 1
                n2 = 0
            else:
                n1 -= 1
                n2 += 1
        return param_list



__author__ = "aymgal"

import numpy as np
import warnings

from lenstronomy.LightModel.Profiles import starlets_util
from lenstronomy.LightModel.Profiles.interpolation import Interpol
from lenstronomy.Util import util

__all__ = ["SLIT_Starlets"]


class SLIT_Starlets(object):
    """Decomposition of an image using the Isotropic Undecimated Walevet Transform, also
    known as "starlet" or "B-spline", using the 'a trous' algorithm.

    Astronomical data (galaxies, stars, ...) are often very sparsely represented in the starlet basis.

    Based on Starck et al. : https://ui.adsabs.harvard.edu/abs/2007ITIP...16..297S/abstract
    """

    param_names = ["amp", "n_scales", "n_pixels", "scale", "center_x", "center_y"]
    param_names_latex = {
        r"$I_0$",
        r"$n_{\rm scales}$",
        r"$n_{\rm pix}$",
        r"scale",
        r"$x_0$",
        r"$y_0$",
    }
    lower_limit_default = {
        "amp": [0],
        "n_scales": 2,
        "n_pixels": 5,
        "center_x": -1000,
        "center_y": -1000,
        "scale": 0.000000001,
    }
    upper_limit_default = {
        "amp": [1e8],
        "n_scales": 20,
        "n_pixels": 1e10,
        "center_x": 1000,
        "center_y": 1000,
        "scale": 10000000000,
    }

    def __init__(
        self,
        thread_count=1,
        fast_inverse=True,
        second_gen=False,
        show_pysap_plots=False,
        force_no_pysap=False,
    ):
        """Load pySAP package if found, and initialize the Starlet transform.

        :param thread_count: number of threads used for pySAP computations
        :param fast_inverse: if True, reconstruction is simply the sum of each scale
            (only for 1st generation starlet transform)
        :param second_gen: if True, uses the second generation of starlet transform
        :param show_pysap_plots: if True, displays pySAP plots when calling the
            decomposition method
        :param force_no_pysap: if True, does not load pySAP and computes starlet
            transforms in python.
        """
        self.use_pysap, pysap = self._load_pysap(force_no_pysap)
        if self.use_pysap:
            self._transf_class = pysap.load_transform(
                "BsplineWaveletTransformATrousAlgorithm"
            )
        else:
            warnings.warn(
                "The python package pySAP is not used for starlet operations. "
                "They will be performed using (slower) python routines."
            )
        self._fast_inverse = fast_inverse
        self._second_gen = second_gen
        self._show_pysap_plots = show_pysap_plots
        self.interpol = Interpol()
        self.thread_count = thread_count

    def function(
        self,
        x,
        y,
        amp=None,
        n_scales=None,
        n_pixels=None,
        scale=1,
        center_x=0,
        center_y=0,
    ):
        """1D inverse starlet transform from starlet coefficients stored in coeffs
        Follows lenstronomy conventions for light profiles.

        :param amp: decomposition coefficients ('amp' to follow conventions in other
            light profile) This is an ndarray with shape (n_scales, sqrt(n_pixels),
            sqrt(n_pixels)) or (n_scales*n_pixels,)
        :param n_scales: number of decomposition scales
        :param n_pixels: number of pixels in a single scale
        :return: reconstructed signal as 1D array of shape (n_pixels,)
        """
        if len(amp.shape) == 1:
            coeffs = util.array2cube(amp, n_scales, n_pixels)
        elif len(amp.shape) == 3:
            coeffs = amp
        else:
            raise ValueError(
                "Starlets 'amp' has not the right shape (1D or 3D arrays are supported)"
            )
        image = self.function_2d(coeffs, n_scales, n_pixels)
        image = self.interpol.function(
            x,
            y,
            image=image,
            scale=scale,
            center_x=center_x,
            center_y=center_y,
            amp=1,
            phi_G=0,
        )
        return image

    def function_2d(self, coeffs, n_scales, n_pixels):
        """2D inverse starlet transform from starlet coefficients stored in coeffs.

        :param coeffs: decomposition coefficients, ndarray with shape (n_scales,
            sqrt(n_pixels), sqrt(n_pixels))
        :param n_scales: number of decomposition scales
        :return: reconstructed signal as 2D array of shape (sqrt(n_pixels),
            sqrt(n_pixels))
        """
        if self.use_pysap and not self._second_gen:
            return self._inverse_transform(coeffs, n_scales, n_pixels)
        else:
            return starlets_util.inverse_transform(
                coeffs, fast=self._fast_inverse, second_gen=self._second_gen
            )

    def decomposition(self, image, n_scales):
        """1D starlet transform from starlet coefficients stored in coeffs.

        :param image: 2D image to be decomposed, ndarray with shape (sqrt(n_pixels),
            sqrt(n_pixels))
        :param n_scales: number of decomposition scales
        :return: reconstructed signal as 1D array of shape (n_scales*n_pixels,)
        """
        if len(image.shape) == 1:
            image_2d = util.array2image(image)
        elif len(image.shape) == 2:
            image_2d = image
        else:
            raise ValueError(
                "image has not the right shape (1D or 2D arrays are supported for starlets decomposition)"
            )
        return util.cube2array(self.decomposition_2d(image_2d, n_scales))

    def decomposition_2d(self, image, n_scales):
        """2D starlet transform from starlet coefficients stored in coeffs.

        :param image: 2D image to be decomposed, ndarray with shape (sqrt(n_pixels),
            sqrt(n_pixels))
        :param n_scales: number of decomposition scales
        :return: reconstructed signal as 2D array of shape (n_scales, sqrt(n_pixels),
            sqrt(n_pixels))
        """
        if self.use_pysap and not self._second_gen:
            coeffs = self._transform(image, n_scales)
        else:
            coeffs = starlets_util.transform(
                image, n_scales, second_gen=self._second_gen
            )
        return coeffs

    def _inverse_transform(self, coeffs, n_scales, n_pixels):
        """Reconstructs image from starlet coefficients."""
        self._check_transform_pysap(n_scales, n_pixels)
        if self._fast_inverse and not self._second_gen:
            # for 1st gen starlet the reconstruction can be performed by summing all scales
            image = np.sum(coeffs, axis=0)
        else:
            coeffs = self._coeffs2pysap(coeffs)
            self._transf.analysis_data = coeffs
            result = self._transf.synthesis()
            if self._show_pysap_plots:
                result.show()
            image = result.data
        return image

    def _transform(self, image, n_scales):
        """Decomposes an image into starlets coefficients."""
        self._check_transform_pysap(n_scales, image.size)
        self._transf.data = image
        self._transf.analysis()
        if self._show_pysap_plots:
            self._transf.show()
        coeffs = self._transf.analysis_data
        coeffs = self._pysap2coeffs(coeffs)
        return coeffs

    def _check_transform_pysap(self, n_scales, n_pixels):
        """If needed, update the loaded pySAP transform to correct number of scales."""
        if (
            not hasattr(self, "_transf")
            or n_scales != self._n_scales
            or n_pixels != self._n_pixels
        ):
            self._transf = self._transf_class(
                nb_scale=n_scales, verbose=False, nb_procs=self.thread_count
            )
            self._n_scales = n_scales
            self._n_pixels = n_pixels

    def _pysap2coeffs(self, coeffs):
        """Convert pySAP decomposition coefficients to numpy array."""
        return np.asarray(coeffs)

    def _coeffs2pysap(self, coeffs):
        """Convert coefficients stored in numpy array to list required by pySAP."""
        coeffs_list = []
        for i in range(coeffs.shape[0]):
            coeffs_list.append(coeffs[i, :, :])
        return coeffs_list

    def _load_pysap(self, force_no_pysap):
        """Load pySAP module."""
        if force_no_pysap:
            return False, None
        try:
            import pysap
        except ImportError:
            return False, None
        else:
            return True, pysap

    def delete_cache(self):
        """Delete the cached interpolated image."""
        self.interpol.delete_cache()



__author__ = "lynevdv"


from lenstronomy.LensModel.Profiles.elliptical_density_slice import ElliSLICE

import numpy as np
import pytest
import numpy.testing as npt


class TestElliSLICE(object):
    """Tests the elliptical slice lens model."""

    def setup_method(self):
        self.ElliSLICE = ElliSLICE()

    def test_function(self):
        x = 0.5
        y = 0.1
        a = 2.0
        b = 1.0
        psi = 30 * np.pi / 180.0
        sigma_0 = 5.0
        values = self.ElliSLICE.function(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal(values, 4.532482297, decimal=4)

        x = 3.0 * np.sqrt(3) / 2.0
        y = 3.0 / 2.0
        values = self.ElliSLICE.function(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal(values, 15.52885056, decimal=4)

        x = np.array([0])
        y = np.array([0])
        values = self.ElliSLICE.function(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal(values[0], 4.054651081, decimal=5)

        x = np.array([np.sqrt(3), np.sqrt(3) + 0.000000001, np.sqrt(3) - 0.000000001])
        y = np.array([1, 1.000000001, 0.999999999])
        values = self.ElliSLICE.function(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal(values[0], values[1], decimal=5)
        npt.assert_almost_equal(values[1], values[2], decimal=5)

    def test_derivatives(self):
        x = 0.5
        y = 0.1
        a = 2.0
        b = 1.0
        psi = 30 * np.pi / 180.0
        sigma_0 = 5.0
        f_x, f_y = self.ElliSLICE.derivatives(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal(f_x, 1.938995765, decimal=6)
        npt.assert_almost_equal(f_y, -0.13835403, decimal=6)

        x = 4
        y = 0.0
        f_x, f_y = self.ElliSLICE.derivatives(x, y, a, b, 0.0, sigma_0)
        npt.assert_almost_equal(f_x, 2.629658164, decimal=6)
        npt.assert_almost_equal(f_y, 0.0, decimal=6)

        x = np.array([0.5])
        y = np.array([0.1])
        f_x, f_y = self.ElliSLICE.derivatives(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal(f_x, 1.938995765, decimal=6)
        npt.assert_almost_equal(f_y, -0.13835403, decimal=6)

        x = np.array([np.sqrt(3), np.sqrt(3) + 0.000000001, np.sqrt(3) - 0.000000001])
        y = np.array([1, 1.000000001, 0.999999999])
        f_x, f_y = self.ElliSLICE.derivatives(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal(f_x[0], f_x[1], decimal=5)
        npt.assert_almost_equal(f_y[1], f_y[2], decimal=5)

    def test_hessian(self):
        x = 0.5
        y = 0.1
        a = 2.0
        b = 1.0
        psi = 30 * np.pi / 180.0
        sigma_0 = 5.0
        f_xx, f_xy, f_yx, f_yy = self.ElliSLICE.hessian(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal((f_xx + f_yy) / 2.0, 5.0, decimal=6)
        x = np.array([1])
        y = np.array([2])
        npt.assert_almost_equal(f_xy, f_yx, decimal=7)
        f_xx, f_xy, f_yx, f_yy = self.ElliSLICE.hessian(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal((f_xx + f_yy) / 2.0, 0.0, decimal=6)
        x = np.array([1, 3, 0.0])
        y = np.array([2, 1, 0.5])
        values = self.ElliSLICE.hessian(x, y, a, b, psi, sigma_0)
        npt.assert_almost_equal((values[0][2] + values[3][2]) / 2.0, 5.0, decimal=6)


if __name__ == "__main__":
    pytest.main()



import numpy as np
import math
import matplotlib.pyplot as plt
import copy

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def sqrt(inputArray, scale_min=None, scale_max=None):
    """Performs sqrt scaling of the input numpy array.

    :type inputArray: numpy array
    :param inputArray: image data array
    :type scale_min: float
    :param scale_min: minimum data value
    :type scale_max: float
    :param scale_max: maximum data value
    :rtype: numpy array
    :return: image data array
    """

    imageData = np.array(inputArray, copy=True)

    if scale_min is None:
        scale_min = imageData.min()
    if scale_max is None:
        scale_max = imageData.max()

    imageData = imageData.clip(min=scale_min, max=scale_max)
    imageData = imageData - scale_min
    indices = np.where(imageData < 0)
    imageData[indices] = 0.0
    imageData = np.sqrt(imageData)
    imageData = imageData / math.sqrt(scale_max - scale_min)
    return imageData


@export
def text_description(
    ax, d, text, color="w", backgroundcolor="k", flipped=False, font_size=15
):
    c_vertical = 1 / 15.0  # + font_size / d / 10.**2
    c_horizontal = 1.0 / 30
    if flipped:
        ax.text(
            d - d * c_horizontal,
            d - d * c_vertical,
            text,
            color=color,
            fontsize=font_size,
            backgroundcolor=backgroundcolor,
        )
    else:
        ax.text(
            d * c_horizontal,
            d - d * c_vertical,
            text,
            color=color,
            fontsize=font_size,
            backgroundcolor=backgroundcolor,
        )


@export
def scale_bar(ax, d, dist=1.0, text='1"', color="w", font_size=15, flipped=False):
    """

    :param ax: matplotlib.axes instance
    :param d: diameter of frame
    :param dist: distance scale printed
    :param text: string printed on scale bar
    :param color: color of scale bar
    :param font_size: font size
    :param flipped: boolean
    :return: None, updated ax instance
    """
    if flipped:
        p0 = d - d / 15.0 - dist
        p1 = d / 15.0
        ax.plot([p0, p0 + dist], [p1, p1], linewidth=2, color=color)
        ax.text(
            p0 + dist / 2.0,
            p1 + 0.01 * d,
            text,
            fontsize=font_size,
            color=color,
            ha="center",
        )
    else:
        p0 = d / 15.0
        ax.plot([p0, p0 + dist], [p0, p0], linewidth=2, color=color)
        ax.text(
            p0 + dist / 2.0,
            p0 + 0.01 * d,
            text,
            fontsize=font_size,
            color=color,
            ha="center",
        )


@export
def coordinate_arrows(ax, d, coords, color="w", font_size=15, arrow_size=0.05):
    """

    :param ax: matplotlib axes instance
    :param d: diameter of frame in ax
    :param coords: lenstronomy.Data.coord_transforms Coordinates() instance
    :param color: color string
    :param font_size: font size of length scale
    :param arrow_size: size of arrow
    :return: updated ax instance
    """
    d0 = d / 8.0
    p0 = d / 15.0
    pt = d / 9.0
    deltaPix = coords.pixel_width
    ra0, dec0 = coords.map_pix2coord((d - d0) / deltaPix, d0 / deltaPix)
    xx_, yy_ = coords.map_coord2pix(ra0, dec0)
    xx_ra, yy_ra = coords.map_coord2pix(ra0 + p0, dec0)
    xx_dec, yy_dec = coords.map_coord2pix(ra0, dec0 + p0)
    xx_ra_t, yy_ra_t = coords.map_coord2pix(ra0 + pt, dec0)
    xx_dec_t, yy_dec_t = coords.map_coord2pix(ra0, dec0 + pt)

    ax.arrow(
        xx_ * deltaPix,
        yy_ * deltaPix,
        (xx_ra - xx_) * deltaPix,
        (yy_ra - yy_) * deltaPix,
        head_width=arrow_size * d,
        head_length=arrow_size * d,
        fc=color,
        ec=color,
        linewidth=1,
    )
    ax.text(
        xx_ra_t * deltaPix,
        yy_ra_t * deltaPix,
        "E",
        color=color,
        fontsize=font_size,
        ha="center",
    )
    ax.arrow(
        xx_ * deltaPix,
        yy_ * deltaPix,
        (xx_dec - xx_) * deltaPix,
        (yy_dec - yy_) * deltaPix,
        head_width=arrow_size * d,
        head_length=arrow_size * d,
        fc=color,
        ec=color,
        linewidth=1,
    )
    ax.text(
        xx_dec_t * deltaPix,
        yy_dec_t * deltaPix,
        "N",
        color=color,
        fontsize=font_size,
        ha="center",
    )


@export
def plot_line_set(
    ax,
    coords,
    line_set_list_x,
    line_set_list_y,
    origin=None,
    flipped_x=False,
    points_only=False,
    pixel_offset=True,
    *args,
    **kwargs
):
    """Plotting a line set on a matplotlib instance where the coordinates are defined in
    pixel units with the lower left corner (defined as origin) is by default (0, 0). The
    coordinates are moved by 0.5 pixels to be placed in the center of the pixel in
    accordance with the matplotlib.matshow() routine.

    :param ax: matplotlib.axis instance
    :param coords: Coordinates() class instance
    :param origin: [x0, y0], lower left pixel coordinate in the frame of the pixels
    :param line_set_list_x: numpy arrays corresponding of different disconnected regions
        of the line (e.g. caustic or critical curve)
    :param line_set_list_y: numpy arrays corresponding of different disconnected regions
        of the line (e.g. caustic or critical curve)
    :param color: string with matplotlib color
    :param flipped_x: bool, if True, flips x-axis
    :param points_only: bool, if True, sets plotting keywords to plot single points
        without connecting lines
    :param pixel_offset: boolean; if True (default plotting), the coordinates are
        shifted a half a pixel to match with the matshow() command to center the
        coordinates in the pixel center
    :return: plot with line sets on matplotlib axis in pixel coordinates
    """
    if origin is None:
        origin = [0, 0]
    pixel_width = coords.pixel_width
    pixel_width_x = pixel_width
    if points_only:
        if "linestyle" not in kwargs:
            kwargs["linestyle"] = ""
        if "marker" not in kwargs:
            kwargs["marker"] = "o"
        if "markersize" not in kwargs:
            kwargs["markersize"] = 0.01
    if flipped_x:
        pixel_width_x = -pixel_width
    if pixel_offset is True:
        shift = 0.5
    else:
        shift = 0
    if isinstance(line_set_list_x, list):
        for i in range(len(line_set_list_x)):
            x_c, y_c = coords.map_coord2pix(line_set_list_x[i], line_set_list_y[i])
            ax.plot(
                (x_c + shift) * pixel_width_x + origin[0],
                (y_c + shift) * pixel_width + origin[1],
                *args,
                **kwargs
            )
    else:
        x_c, y_c = coords.map_coord2pix(line_set_list_x, line_set_list_y)
        ax.plot(
            (x_c + shift) * pixel_width_x + origin[0],
            (y_c + shift) * pixel_width + origin[1],
            *args,
            **kwargs
        )
    return ax


@export
def image_position_plot(
    ax,
    coords,
    ra_image,
    dec_image,
    color="w",
    image_name_list=None,
    origin=None,
    flipped_x=False,
    pixel_offset=True,
    plot_out_of_image=True,
):
    """

    :param ax: matplotlib axis instance
    :param coords: Coordinates() class instance or inherited class (such as PixelGrid(), or Data())
    :param ra_image: Ra/x-coordinates of image positions (list of arrays in angular units)
    :param dec_image: Dec/y-coordinates of image positions (list of arrays in angular units)
    :param color: color of ticks and text
    :param image_name_list: list of strings for names of the images in the same order as the positions
    :param origin: [x0, y0], lower left pixel coordinate in the frame of the pixels
    :param flipped_x: bool, if True, flips x-axis
    :param pixel_offset: boolean; if True (default plotting), the coordinates are shifted a half a pixel to match with
     the matshow() command to center the coordinates in the pixel center
    :param plot_out_of_image: if True, plots images even appearing out of the Coordinate frame
    :type plot_out_of_image: bool
    :return: matplotlib axis instance with images plotted on
    """
    if origin is None:
        origin = [0, 0]
    pixel_width = coords.pixel_width
    pixel_width_x = pixel_width
    if flipped_x:
        pixel_width_x = -pixel_width
    if image_name_list is None:
        image_name_list = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L"]

    if not isinstance(ra_image, list):
        ra_image_, dec_image_ = [ra_image], [dec_image]
    else:
        ra_image_, dec_image_ = ra_image, dec_image
    if pixel_offset is True:
        shift = 0.5
    else:
        shift = 0
    try:
        nx, ny = coords.num_pixel_axes
    except:
        plot_out_of_image = True
    for ra, dec in zip(ra_image_, dec_image_):
        x_image, y_image = coords.map_coord2pix(ra, dec)

        for i in range(len(x_image)):
            if not plot_out_of_image:
                if 0 < x_image[i] < nx and 0 < y_image[i] < ny:
                    x_ = (x_image[i] + shift) * pixel_width_x + origin[0]
                    y_ = (y_image[i] + shift) * pixel_width + origin[1]
                    ax.plot(x_, y_, "o", color=color)
                    ax.text(x_, y_, image_name_list[i], fontsize=20, color=color)
    return ax


@export
def source_position_plot(
    ax, coords, ra_source, dec_source, marker="*", markersize=10, **kwargs
):
    """

    :param ax: matplotlib axis instance
    :param coords: Coordinates() class instance or inherited class (such as PixelGrid(), or Data())
    :param ra_source: list of source position in angular units
    :param dec_source: list of source position in angular units
    :param marker: marker style for matplotlib
    :param markersize: marker size for matplotlib
    :return: matplotlib axis instance with images plotted on
    """
    delta_pix = coords.pixel_width
    if len(ra_source) > 0:
        for ra, dec in zip(ra_source, dec_source):
            x_source, y_source = coords.map_coord2pix(ra, dec)
            ax.plot(
                (x_source + 0.5) * delta_pix,
                (y_source + 0.5) * delta_pix,
                marker=marker,
                markersize=markersize,
                **kwargs
            )
    return ax


@export
def result_string(x, weights=None, title_fmt=".2f", label=None):
    """

    :param x: marginalized 1-d posterior
    :param weights: weights of posteriors (optional)
    :param title_fmt: format to what digit the results are presented
    :param label: string of parameter label (optional)
    :return: string with mean :math:`\\pm` quartile
    """
    from corner import quantile

    q_16, q_50, q_84 = quantile(x, [0.16, 0.5, 0.84], weights=weights)
    q_m, q_p = q_50 - q_16, q_84 - q_50

    # Format the quantile display.
    fmt = "{{0:{0}}}".format(title_fmt).format
    title = r"${{{0}}}_{{-{1}}}^{{+{2}}}$"
    title = title.format(fmt(q_50), fmt(q_m), fmt(q_p))
    if label is not None:
        title = "{0} = {1}".format(label, title)
    return title


@export
def cmap_conf(cmap_string):
    """Configures matplotlib color map.

    :param cmap_string: string of cmap name, or cmap instance
    :return: cmap instance with setting for bad pixels and values below the threshold
    """
    if isinstance(cmap_string, str):
        cmap = plt.get_cmap(cmap_string)
    else:
        cmap = cmap_string
    # cmap_new = cmap.copy()
    cmap_new = copy.deepcopy(cmap)
    cmap_new.set_bad(color="k", alpha=1.0)
    cmap_new.set_under("k")
    return cmap_new
